import logging

import networkx as nx

from cubed.primitive.blockwise import (
    can_fuse_multiple_primitive_ops,
    can_fuse_primitive_ops,
    fuse,
    fuse_multiple,
)

logger = logging.getLogger(__name__)


def simple_optimize_dag(dag):
    """Apply map blocks fusion."""

    # note there is no need to prune the dag, since the way it is built
    # ensures that only the transitive dependencies of the target arrays are included

    dag = dag.copy()
    nodes = {n: d for (n, d) in dag.nodes(data=True)}

    def can_fuse(n):
        # fuse a single chain looking like this:
        # op1 -> op2_input -> op2

        op2 = n

        # if node (op2) does not have a primitive op then it can't be fused
        if "primitive_op" not in nodes[op2]:
            return False

        # if node (op2) does not have exactly one input then don't fuse
        # (it could have no inputs or multiple inputs)
        if dag.in_degree(op2) != 1:
            return False

        # if input is used by another node then don't fuse
        op2_input = next(dag.predecessors(op2))
        if dag.out_degree(op2_input) != 1:
            return False

        # if node producing input (op1) has more than one output then don't fuse
        op1 = next(dag.predecessors(op2_input))
        if dag.out_degree(op1) != 1:
            return False

        # op1 and op2 must have primitive ops that can be fused
        if "primitive_op" not in nodes[op1]:
            return False
        return can_fuse_primitive_ops(
            nodes[op1]["primitive_op"], nodes[op2]["primitive_op"]
        )

    for n in list(dag.nodes()):
        if can_fuse(n):
            op2 = n
            op2_input = next(dag.predecessors(op2))
            op1 = next(dag.predecessors(op2_input))
            op1_inputs = list(dag.predecessors(op1))

            primitive_op = fuse(nodes[op1]["primitive_op"], nodes[op2]["primitive_op"])
            nodes[op2]["primitive_op"] = primitive_op
            nodes[op2]["pipeline"] = primitive_op.pipeline

            for n in op1_inputs:
                dag.add_edge(n, op2)
            dag.remove_node(op2_input)
            dag.remove_node(op1)

    return dag


sym_counter = 0


def gensym(name="op"):
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


def predecessors_unordered(dag, name):
    """Return a node's predecessors in no particular order, with repeats for multiple edges."""
    for pre, _ in dag.in_edges(name):
        yield pre


def predecessor_ops(dag, name):
    """Return an op node's op predecessors in the same order as the input source arrays for the op.

    Note that each input source array is produced by a single predecessor op.
    """
    nodes = dict(dag.nodes(data=True))
    for input in nodes[name]["primitive_op"].source_array_names:
        pre_list = list(predecessors_unordered(dag, input))
        assert len(pre_list) == 1  # each array is produced by a single op
        yield pre_list[0]


def is_fusable(node_dict):
    "Return True if a node can be fused."
    return "primitive_op" in node_dict and node_dict["primitive_op"].fusable


def can_fuse_predecessors(
    dag,
    name,
    *,
    max_total_source_arrays=4,
    max_total_num_input_blocks=None,
    always_fuse=None,
    never_fuse=None,
):
    nodes = dict(dag.nodes(data=True))

    # if node itself can't be fused then there is nothing to fuse
    if not is_fusable(nodes[name]):
        logger.debug(
            "can't fuse %s since it is not a primitive operation, or it uses map_direct",
            name,
        )
        return False

    # if no predecessor ops can be fused then there is nothing to fuse
    if all(not is_fusable(nodes[pre]) for pre in predecessor_ops(dag, name)):
        logger.debug("can't fuse %s since no predecessor ops can be fused", name)
        return False

    # if node is in never_fuse or always_fuse list then it overrides logic below
    if never_fuse is not None and name in never_fuse:
        logger.debug("can't fuse %s since it is in 'never_fuse'", name)
        return False
    if always_fuse is not None and name in always_fuse:
        logger.debug("can fuse %s since it is in 'always_fuse'", name)
        return True

    # if there is more than a single predecessor op, and the total number of source arrays to
    # the fused function would be more than an allowed maximum, then don't fuse
    if len(list(predecessor_ops(dag, name))) > 1:
        total_source_arrays = sum(
            len(list(predecessors_unordered(dag, pre))) if is_fusable(nodes[pre]) else 1
            for pre in predecessor_ops(dag, name)
        )
        if total_source_arrays > max_total_source_arrays:
            logger.debug(
                "can't fuse %s since total number of source arrays (%s) exceeds max (%s)",
                name,
                total_source_arrays,
                max_total_source_arrays,
            )
            return False

    predecessor_primitive_ops = [
        nodes[pre]["primitive_op"]
        for pre in predecessor_ops(dag, name)
        if is_fusable(nodes[pre])
    ]
    return can_fuse_multiple_primitive_ops(
        name,
        nodes[name]["primitive_op"],
        predecessor_primitive_ops,
        max_total_num_input_blocks=max_total_num_input_blocks,
    )


def fuse_predecessors(
    dag,
    name,
    *,
    max_total_source_arrays=4,
    max_total_num_input_blocks=None,
    always_fuse=None,
    never_fuse=None,
):
    """Fuse a node with its immediate predecessors."""

    # if can't fuse then return dag unchanged
    if not can_fuse_predecessors(
        dag,
        name,
        max_total_source_arrays=max_total_source_arrays,
        max_total_num_input_blocks=max_total_num_input_blocks,
        always_fuse=always_fuse,
        never_fuse=never_fuse,
    ):
        return dag

    nodes = dict(dag.nodes(data=True))

    primitive_op = nodes[name]["primitive_op"]

    # if a predecessor has no primitive op then just use None
    predecessor_primitive_ops = [
        nodes[pre]["primitive_op"] if is_fusable(nodes[pre]) else None
        for pre in predecessor_ops(dag, name)
    ]

    fused_primitive_op = fuse_multiple(primitive_op, *predecessor_primitive_ops)

    fused_dag = dag.copy()
    fused_nodes = dict(fused_dag.nodes(data=True))

    fused_nodes[name]["primitive_op"] = fused_primitive_op
    fused_nodes[name]["pipeline"] = fused_primitive_op.pipeline

    # re-wire dag to remove predecessor nodes that have been fused

    # 1. update edges to change inputs
    for input in predecessors_unordered(dag, name):
        pre = next(predecessors_unordered(dag, input))
        if not is_fusable(fused_nodes[pre]):
            # if a predecessor is not fusable then don't change the edge
            continue
        fused_dag.remove_edge(input, name)
    for pre in predecessor_ops(dag, name):
        if not is_fusable(fused_nodes[pre]):
            # if a predecessor is not fusable then don't change the edge
            continue
        for input in predecessors_unordered(dag, pre):
            fused_dag.add_edge(input, name)

    # 2. remove predecessor nodes with no successors
    # (ones with successors are needed by other nodes)
    for input in predecessors_unordered(dag, name):
        if fused_dag.out_degree(input) == 0:
            for pre in list(predecessors_unordered(fused_dag, input)):
                fused_dag.remove_node(pre)
            fused_dag.remove_node(input)

    return fused_dag


def multiple_inputs_optimize_dag(
    dag,
    *,
    max_total_source_arrays=4,
    max_total_num_input_blocks=None,
    always_fuse=None,
    never_fuse=None,
):
    """Fuse multiple inputs."""
    for name in list(nx.topological_sort(dag)):
        if name.startswith("array-"):
            continue
        dag = fuse_predecessors(
            dag,
            name,
            max_total_source_arrays=max_total_source_arrays,
            max_total_num_input_blocks=max_total_num_input_blocks,
            always_fuse=always_fuse,
            never_fuse=never_fuse,
        )
    return dag


def fuse_all_optimize_dag(dag):
    """Force all operations to be fused."""
    dag = dag.copy()
    always_fuse = [op for op in dag.nodes() if op.startswith("op-")]
    return multiple_inputs_optimize_dag(dag, always_fuse=always_fuse)


def fuse_only_optimize_dag(dag, *, only_fuse=None):
    """Force only specified operations to be fused, all others will be left even if they are suitable for fusion."""
    dag = dag.copy()
    always_fuse = only_fuse
    never_fuse = set(op for op in dag.nodes() if op.startswith("op-")) - set(only_fuse)
    return multiple_inputs_optimize_dag(
        dag, always_fuse=always_fuse, never_fuse=never_fuse
    )



import asyncio
import time
from asyncio.exceptions import TimeoutError
from typing import Any, AsyncIterator, Iterable, Optional, Sequence

from aiostream import stream
from modal.exception import ConnectionError
from modal.functions import Function
from networkx import MultiDiGraph
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from cubed.runtime.executors.asyncio import async_map_unordered
from cubed.runtime.executors.modal import (
    Container,
    check_runtime_memory,
    run_remotely,
    stub,
)
from cubed.runtime.pipeline import visit_node_generations, visit_nodes
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import handle_callbacks, handle_operation_start_callbacks
from cubed.spec import Spec


# We need map_unordered for the use_backups implementation
async def map_unordered(
    app_function: Function,
    input: Iterable[Any],
    use_backups: bool = True,
    backup_function: Optional[Function] = None,
    batch_size: Optional[int] = None,
    return_stats: bool = False,
    name: Optional[str] = None,
    **kwargs,
) -> AsyncIterator[Any]:
    """
    Apply a function to items of an input list, yielding results as they are completed
    (which may be different to the input order).

    :param app_function: The Modal function to map over the data.
    :param input: An iterable of input data.
    :param use_backups: Whether to launch backup tasks to mitigate against slow-running tasks.
    :param kwargs: Keyword arguments to pass to the function.

    :return: Function values (and optionally stats) as they are completed, not necessarily in the input order.
    """

    if not use_backups and batch_size is None:
        task_create_tstamp = time.time()
        async for result in app_function.map(input, order_outputs=False, kwargs=kwargs):
            if return_stats:
                result, stats = result
                if name is not None:
                    stats["name"] = name
                stats["task_create_tstamp"] = task_create_tstamp
                yield result, stats
            else:
                yield result
        return

    def create_futures_func(input, **kwargs):
        return [
            (i, asyncio.ensure_future(app_function.remote.aio(i, **kwargs)))
            for i in input
        ]

    backup_function = backup_function or app_function

    def create_backup_futures_func(input, **kwargs):
        return [
            (i, asyncio.ensure_future(backup_function.remote.aio(i, **kwargs)))
            for i in input
        ]

    async for result in async_map_unordered(
        create_futures_func,
        input,
        use_backups=use_backups,
        create_backup_futures_func=create_backup_futures_func,
        batch_size=batch_size,
        return_stats=return_stats,
        name=name,
        **kwargs,
    ):
        yield result


def pipeline_to_stream(app_function, name, pipeline, **kwargs):
    return stream.iterate(
        map_unordered(
            app_function,
            pipeline.mappable,
            return_stats=True,
            name=name,
            func=pipeline.function,
            config=pipeline.config,
            **kwargs,
        )
    )


# This just retries the initial connection attempt, not the function calls
@retry(
    reraise=True,
    retry=retry_if_exception_type((TimeoutError, ConnectionError)),
    stop=stop_after_attempt(3),
)
async def async_execute_dag(
    dag: MultiDiGraph,
    callbacks: Optional[Sequence[Callback]] = None,
    resume: Optional[bool] = None,
    spec: Optional[Spec] = None,
    cloud: Optional[str] = None,
    compute_arrays_in_parallel: Optional[bool] = None,
    **kwargs,
) -> None:
    if spec is not None:
        check_runtime_memory(spec)
    async with stub.run():
        cloud = cloud or "aws"
        if cloud == "aws":
            app_function = run_remotely
        elif cloud == "gcp":
            app_function = Container().run_remotely
        else:
            raise ValueError(f"Unrecognized cloud: {cloud}")
        if not compute_arrays_in_parallel:
            # run one pipeline at a time
            for name, node in visit_nodes(dag, resume=resume):
                handle_operation_start_callbacks(callbacks, name)
                st = pipeline_to_stream(app_function, name, node["pipeline"], **kwargs)
                async with st.stream() as streamer:
                    async for _, stats in streamer:
                        handle_callbacks(callbacks, stats)
        else:
            for gen in visit_node_generations(dag, resume=resume):
                # run pipelines in the same topological generation in parallel by merging their streams
                streams = [
                    pipeline_to_stream(app_function, name, node["pipeline"], **kwargs)
                    for name, node in gen
                ]
                merged_stream = stream.merge(*streams)
                async with merged_stream.stream() as streamer:
                    async for _, stats in streamer:
                        handle_callbacks(callbacks, stats)


class AsyncModalDagExecutor(DagExecutor):
    """An execution engine that uses Modal's async API."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "modal"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        resume: Optional[bool] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        asyncio.run(
            async_execute_dag(
                dag,
                callbacks=callbacks,
                resume=resume,
                spec=spec,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )



from importlib.metadata import version as _version

try:
    __version__ = _version("cubed")
except Exception:  # pragma: no cover
    __version__ = "unknown"

from donfig import Config

config = Config(
    "cubed",
    # default spec is local temp dir and a modest amount of memory (200MB, of which 100MB is reserved)
    defaults=[{"spec": {"allowed_mem": 200_000_000, "reserved_mem": 100_000_000}}],
)

from .array_api import Array
from .core.array import compute, measure_reserved_mem, visualize
from .core.gufunc import apply_gufunc
from .core.ops import from_array, from_zarr, map_blocks, store, to_zarr
from .nan_functions import nanmean, nansum
from .runtime.types import Callback, TaskEndEvent
from .spec import Spec

__all__ = [
    "__version__",
    "Callback",
    "Array",
    "Spec",
    "TaskEndEvent",
    "apply_gufunc",
    "compute",
    "config",
    "from_array",
    "from_zarr",
    "map_blocks",
    "measure_reserved_mem",
    "nanmean",
    "nansum",
    "store",
    "to_zarr",
    "visualize",
]



from dataclasses import asdict
from pathlib import Path

import pandas as pd

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback


class HistoryCallback(Callback):
    def on_compute_start(self, event):
        plan = []
        for name, node in visit_nodes(event.dag, event.resume):
            primitive_op = node["primitive_op"]
            plan.append(
                dict(
                    name=name,
                    op_name=node["op_name"],
                    projected_mem=primitive_op.projected_mem,
                    reserved_mem=primitive_op.reserved_mem,
                    num_tasks=primitive_op.num_tasks,
                )
            )

        self.plan = plan
        self.events = []

    def on_task_end(self, event):
        self.events.append(asdict(event))

    def on_compute_end(self, event):
        self.plan_df = pd.DataFrame(self.plan)
        self.events_df = pd.DataFrame(self.events)
        history_path = Path(f"history/{event.compute_id}")
        history_path.mkdir(parents=True, exist_ok=True)
        self.plan_df_path = history_path / "plan.csv"
        self.events_df_path = history_path / "events.csv"
        self.stats_df_path = history_path / "stats.csv"
        self.plan_df.to_csv(self.plan_df_path, index=False)
        self.events_df.to_csv(self.events_df_path, index=False)

        self.stats_df = analyze(self.plan_df, self.events_df)
        self.stats_df.to_csv(self.stats_df_path, index=False)


def analyze(plan_df, events_df):
    # convert memory to MB
    plan_df["projected_mem_mb"] = plan_df["projected_mem"] / 1_000_000
    plan_df["reserved_mem_mb"] = plan_df["reserved_mem"] / 1_000_000
    plan_df = plan_df[
        [
            "name",
            "op_name",
            "projected_mem_mb",
            "reserved_mem_mb",
            "num_tasks",
        ]
    ]
    events_df["peak_measured_mem_start_mb"] = (
        events_df["peak_measured_mem_start"] / 1_000_000
    )
    events_df["peak_measured_mem_end_mb"] = (
        events_df["peak_measured_mem_end"] / 1_000_000
    )
    events_df["peak_measured_mem_delta_mb"] = (
        events_df["peak_measured_mem_end_mb"] - events_df["peak_measured_mem_start_mb"]
    )

    # find per-array stats
    df = events_df.groupby("name", as_index=False).agg(
        {
            "peak_measured_mem_start_mb": ["min", "mean", "max"],
            "peak_measured_mem_end_mb": ["max"],
            "peak_measured_mem_delta_mb": ["min", "mean", "max"],
        }
    )

    # flatten multi-index
    df.columns = ["_".join(a).rstrip("_") for a in df.columns.to_flat_index()]
    df = df.merge(plan_df, on="name")

    def projected_mem_utilization(row):
        return row["peak_measured_mem_end_mb_max"] / row["projected_mem_mb"]

    df["projected_mem_utilization"] = df.apply(
        lambda row: projected_mem_utilization(row), axis=1
    )
    df = df[
        [
            "name",
            "op_name",
            "num_tasks",
            "peak_measured_mem_start_mb_max",
            "peak_measured_mem_end_mb_max",
            "peak_measured_mem_delta_mb_max",
            "projected_mem_mb",
            "reserved_mem_mb",
            "projected_mem_utilization",
        ]
    ]

    return df



import pytest

from cubed.array_api import ones


# This test is adapted from dask, see the license in cubed/vendor/dask/LICENSE.txt
def test_repr_html():
    pytest.importorskip("jinja2")
    assert ones([])._repr_html_()
    assert ones(10)[:0]._repr_html_()
    assert ones(10)._repr_html_()
    assert ones((10, 10))._repr_html_()
    assert ones((10, 10, 10))._repr_html_()
    assert ones((10, 10, 10, 10))._repr_html_()



import inspect
import itertools
import platform

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from cubed.backend_array_api import namespace as nxp
from cubed.utils import (
    array_memory,
    block_id_to_offset,
    broadcast_trick,
    extract_stack_summaries,
    join_path,
    map_nested,
    memory_repr,
    offset_to_block_id,
    peak_measured_mem,
    split_into,
    to_chunksize,
)


def test_array_memory():
    assert array_memory(np.int64, (3,)) == 24
    assert array_memory(np.int32, (3,)) == 12
    assert array_memory(np.int32, (3, 5)) == 60
    assert array_memory(np.int32, (0,)) == 0


def test_block_id_to_offset():
    numblocks = (5, 3)
    for block_id in itertools.product(*[list(range(n)) for n in numblocks]):
        offset = block_id_to_offset(block_id, numblocks)
        assert offset_to_block_id(offset, numblocks) == block_id

    with pytest.raises(ValueError):
        block_id_to_offset((6, 12), numblocks)

    with pytest.raises(ValueError):
        offset_to_block_id(100, numblocks)


def test_to_chunksize():
    assert to_chunksize(((3, 3, 3, 1),)) == (3,)
    assert to_chunksize(((0,),)) == (1,)  # Zarr doesn't support zero-length chunks
    with pytest.raises(ValueError):
        to_chunksize(((3, 2, 3, 3, 1),))


def test_join_path():
    assert join_path("http://host/path", "subpath") == "http://host/path/subpath"
    assert join_path("http://host/path/", "subpath") == "http://host/path/subpath"
    assert (
        join_path("http://host/path?a=b", "subpath") == "http://host/path/subpath?a=b"
    )
    assert (
        join_path("http://host/path/?a=b", "subpath") == "http://host/path/subpath?a=b"
    )
    assert join_path("http://host/path#a", "subpath") == "http://host/path/subpath#a"
    assert join_path("s3://host/path", "subpath") == "s3://host/path/subpath"
    assert join_path("relative_path/path", "subpath") == "relative_path/path/subpath"
    assert join_path("/absolute_path/path", "subpath") == "/absolute_path/path/subpath"
    assert (
        join_path("http://host/a%20path", "subpath") == "http://host/a%20path/subpath"
    )
    assert join_path("http://host/a path", "subpath") == "http://host/a%20path/subpath"


def test_memory_repr():
    assert memory_repr(0) == "0 bytes"
    assert memory_repr(1) == "1 bytes"
    assert memory_repr(999) == "999 bytes"
    assert memory_repr(1_000) == "1.0 KB"
    assert memory_repr(9_999) == "10.0 KB"
    assert memory_repr(1_000_000) == "1.0 MB"
    assert memory_repr(1_000_000_000_000_000) == "1.0 PB"
    assert memory_repr(int(1e18)) == "1.0e+18 bytes"
    with pytest.raises(ValueError):
        memory_repr(-1)


@pytest.mark.skipif(platform.system() == "Windows", reason="does not run on windows")
def test_peak_measured_mem():
    assert peak_measured_mem() > 0


def test_extract_stack_summaries():
    frame = inspect.currentframe()
    stack_summaries = extract_stack_summaries(frame)
    assert stack_summaries[-1].name == "test_extract_stack_summaries"
    assert stack_summaries[-1].module == "cubed.tests.test_utils"
    assert not stack_summaries[-1].is_cubed()


def test_split_into():
    assert list(split_into([1, 2, 3, 4, 5, 6], [1, 2, 3])) == [[1], [2, 3], [4, 5, 6]]
    assert list(split_into([1, 2, 3, 4, 5, 6], [2, 3])) == [[1, 2], [3, 4, 5]]
    assert list(split_into([1, 2, 3, 4], [1, 2, 3, 4])) == [[1], [2, 3], [4], []]


def test_map_nested_lists():
    inc = lambda x: x + 1

    assert map_nested(inc, [1, 2]) == [2, 3]
    assert map_nested(inc, [[1, 2]]) == [[2, 3]]
    assert map_nested(inc, [[1, 2], [3, 4]]) == [[2, 3], [4, 5]]


count = 0


def inc(x):
    global count
    count = count + 1
    return x + 1


def test_map_nested_iterators():
    # same tests as test_map_nested_lists, but use a counter to check that iterators are advanced at correct points
    global count

    out = map_nested(inc, iter([1, 2]))
    assert isinstance(out, map)
    assert count == 0
    assert list(out) == [2, 3]
    assert count == 2

    # reset count
    count = 0

    out = map_nested(inc, [iter([1, 2])])
    assert isinstance(out, list)
    assert count == 0
    assert len(out) == 1
    out = out[0]
    assert isinstance(out, map)
    assert count == 0
    assert list(out) == [2, 3]
    assert count == 2

    # reset count
    count = 0

    out = map_nested(inc, [iter([1, 2]), iter([3, 4])])
    assert isinstance(out, list)
    assert count == 0
    assert len(out) == 2
    out0 = out[0]
    assert isinstance(out0, map)
    assert count == 0
    assert list(out0) == [2, 3]
    assert count == 2
    out1 = out[1]
    assert isinstance(out1, map)
    assert count == 2
    assert list(out1) == [4, 5]
    assert count == 4


def test_broadcast_trick():
    a = nxp.ones((10, 10), dtype=nxp.int8)
    b = broadcast_trick(nxp.ones)((10, 10), dtype=nxp.int8)

    assert_array_equal(a, b)
    assert a.nbytes == 100
    assert b.base.nbytes == 1

    a = nxp.ones((), dtype=nxp.int8)
    b = broadcast_trick(nxp.ones)((), dtype=nxp.int8)
    assert_array_equal(a, b)



import numbers

import numpy as np


def validate_axis(axis, ndim):
    """Validate an input to axis= keywords"""
    if isinstance(axis, (tuple, list)):
        return tuple(validate_axis(ax, ndim) for ax in axis)
    if not isinstance(axis, numbers.Integral):
        raise TypeError("Axis value must be an integer, got %s" % axis)
    if axis < -ndim or axis >= ndim:
        raise np.AxisError(
            "Axis %d is out of bounds for array of dimension %d" % (axis, ndim)
        )
    if axis < 0:
        axis += ndim
    return axis



from cubed.array_api.data_type_functions import result_type
from cubed.array_api.dtypes import (
    _boolean_dtypes,
    _complex_floating_dtypes,
    _floating_dtypes,
    _integer_dtypes,
    _integer_or_boolean_dtypes,
    _numeric_dtypes,
    _real_floating_dtypes,
    _real_numeric_dtypes,
    complex64,
    complex128,
    float32,
    float64,
)
from cubed.backend_array_api import namespace as nxp
from cubed.core import elemwise


def abs(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in abs")
    if x.dtype == complex64:
        dtype = float32
    elif x.dtype == complex128:
        dtype = float64
    else:
        dtype = x.dtype
    return elemwise(nxp.abs, x, dtype=dtype)


def acos(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in acos")
    return elemwise(nxp.acos, x, dtype=x.dtype)


def acosh(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in acosh")
    return elemwise(nxp.acosh, x, dtype=x.dtype)


def add(x1, x2, /):
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in add")
    return elemwise(nxp.add, x1, x2, dtype=result_type(x1, x2))


def asin(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in asin")
    return elemwise(nxp.asin, x, dtype=x.dtype)


def asinh(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in asinh")
    return elemwise(nxp.asinh, x, dtype=x.dtype)


def atan(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in atan")
    return elemwise(nxp.atan, x, dtype=x.dtype)


def atan2(x1, x2, /):
    if x1.dtype not in _real_floating_dtypes or x2.dtype not in _real_floating_dtypes:
        raise TypeError("Only real floating-point dtypes are allowed in atan2")
    return elemwise(nxp.atan2, x1, x2, dtype=result_type(x1, x2))


def atanh(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in atanh")
    return elemwise(nxp.atanh, x, dtype=x.dtype)


def bitwise_and(x1, x2, /):
    if (
        x1.dtype not in _integer_or_boolean_dtypes
        or x2.dtype not in _integer_or_boolean_dtypes
    ):
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_and")
    return elemwise(nxp.bitwise_and, x1, x2, dtype=result_type(x1, x2))


def bitwise_invert(x, /):
    if x.dtype not in _integer_or_boolean_dtypes:
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_invert")
    return elemwise(nxp.bitwise_invert, x, dtype=x.dtype)


def bitwise_left_shift(x1, x2, /):
    if x1.dtype not in _integer_dtypes or x2.dtype not in _integer_dtypes:
        raise TypeError("Only integer dtypes are allowed in bitwise_left_shift")
    return elemwise(nxp.bitwise_left_shift, x1, x2, dtype=result_type(x1, x2))


def bitwise_or(x1, x2, /):
    if (
        x1.dtype not in _integer_or_boolean_dtypes
        or x2.dtype not in _integer_or_boolean_dtypes
    ):
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_or")
    return elemwise(nxp.bitwise_or, x1, x2, dtype=result_type(x1, x2))


def bitwise_right_shift(x1, x2, /):
    if x1.dtype not in _integer_dtypes or x2.dtype not in _integer_dtypes:
        raise TypeError("Only integer dtypes are allowed in bitwise_right_shift")
    return elemwise(nxp.bitwise_right_shift, x1, x2, dtype=result_type(x1, x2))


def bitwise_xor(x1, x2, /):
    if (
        x1.dtype not in _integer_or_boolean_dtypes
        or x2.dtype not in _integer_or_boolean_dtypes
    ):
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_xor")
    return elemwise(nxp.bitwise_xor, x1, x2, dtype=result_type(x1, x2))


def ceil(x, /):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in ceil")
    if x.dtype in _integer_dtypes:
        # Note: The return dtype of ceil is the same as the input
        return x
    return elemwise(nxp.ceil, x, dtype=x.dtype)


def conj(x, /):
    if x.dtype not in _complex_floating_dtypes:
        raise TypeError("Only complex floating-point dtypes are allowed in conj")
    return elemwise(nxp.conj, x, dtype=x.dtype)


def cos(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in cos")
    return elemwise(nxp.cos, x, dtype=x.dtype)


def cosh(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in cosh")
    return elemwise(nxp.cosh, x, dtype=x.dtype)


def divide(x1, x2, /):
    if x1.dtype not in _floating_dtypes or x2.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in divide")
    return elemwise(nxp.divide, x1, x2, dtype=result_type(x1, x2))


def exp(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in exp")
    return elemwise(nxp.exp, x, dtype=x.dtype)


def expm1(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in expm1")
    return elemwise(nxp.expm1, x, dtype=x.dtype)


def equal(x1, x2, /):
    return elemwise(nxp.equal, x1, x2, dtype=nxp.bool)


def floor(x, /):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in floor")
    if x.dtype in _integer_dtypes:
        # Note: The return dtype of floor is the same as the input
        return x
    return elemwise(nxp.floor, x, dtype=x.dtype)


def floor_divide(x1, x2, /):
    if x1.dtype not in _real_numeric_dtypes or x2.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in floor_divide")
    return elemwise(nxp.floor_divide, x1, x2, dtype=result_type(x1, x2))


def greater(x1, x2, /):
    return elemwise(nxp.greater, x1, x2, dtype=nxp.bool)


def greater_equal(x1, x2, /):
    return elemwise(nxp.greater_equal, x1, x2, dtype=nxp.bool)


def imag(x, /):
    if x.dtype == complex64:
        dtype = float32
    elif x.dtype == complex128:
        dtype = float64
    else:
        raise TypeError("Only complex floating-point dtypes are allowed in imag")
    return elemwise(nxp.imag, x, dtype=dtype)


def isfinite(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in isfinite")
    return elemwise(nxp.isfinite, x, dtype=nxp.bool)


def isinf(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in isinf")
    return elemwise(nxp.isinf, x, dtype=nxp.bool)


def isnan(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in isnan")
    return elemwise(nxp.isnan, x, dtype=nxp.bool)


def less(x1, x2, /):
    return elemwise(nxp.less, x1, x2, dtype=nxp.bool)


def less_equal(x1, x2, /):
    return elemwise(nxp.less_equal, x1, x2, dtype=nxp.bool)


def log(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log")
    return elemwise(nxp.log, x, dtype=x.dtype)


def log1p(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log1p")
    return elemwise(nxp.log1p, x, dtype=x.dtype)


def log2(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log2")
    return elemwise(nxp.log2, x, dtype=x.dtype)


def log10(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log10")
    return elemwise(nxp.log10, x, dtype=x.dtype)


def logaddexp(x1, x2, /):
    if x1.dtype not in _real_floating_dtypes or x2.dtype not in _real_floating_dtypes:
        raise TypeError("Only real floating-point dtypes are allowed in logaddexp")
    return elemwise(nxp.logaddexp, x1, x2, dtype=result_type(x1, x2))


def logical_and(x1, x2, /):
    if x1.dtype not in _boolean_dtypes or x2.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_and")
    return elemwise(nxp.logical_and, x1, x2, dtype=nxp.bool)


def logical_not(x, /):
    if x.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_not")
    return elemwise(nxp.logical_not, x, dtype=nxp.bool)


def logical_or(x1, x2, /):
    if x1.dtype not in _boolean_dtypes or x2.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_or")
    return elemwise(nxp.logical_or, x1, x2, dtype=nxp.bool)


def logical_xor(x1, x2, /):
    if x1.dtype not in _boolean_dtypes or x2.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_xor")
    return elemwise(nxp.logical_xor, x1, x2, dtype=nxp.bool)


def multiply(x1, x2, /):
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in multiply")
    return elemwise(nxp.multiply, x1, x2, dtype=result_type(x1, x2))


def negative(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in negative")
    return elemwise(nxp.negative, x, dtype=x.dtype)


def not_equal(x1, x2, /):
    return elemwise(nxp.not_equal, x1, x2, dtype=nxp.bool)


def positive(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in positive")
    return elemwise(nxp.positive, x, dtype=x.dtype)


def pow(x1, x2, /):
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in pow")
    return elemwise(nxp.pow, x1, x2, dtype=result_type(x1, x2))


def real(x, /):
    if x.dtype == complex64:
        dtype = float32
    elif x.dtype == complex128:
        dtype = float64
    else:
        raise TypeError("Only complex floating-point dtypes are allowed in real")
    return elemwise(nxp.real, x, dtype=dtype)


def remainder(x1, x2, /):
    if x1.dtype not in _real_numeric_dtypes or x2.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in remainder")
    return elemwise(nxp.remainder, x1, x2, dtype=result_type(x1, x2))


def round(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in round")
    return elemwise(nxp.round, x, dtype=x.dtype)


def sign(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in sign")
    return elemwise(nxp.sign, x, dtype=x.dtype)


def sin(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in sin")
    return elemwise(nxp.sin, x, dtype=x.dtype)


def sinh(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in sinh")
    return elemwise(nxp.sinh, x, dtype=x.dtype)


def sqrt(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in sqrt")
    return elemwise(nxp.sqrt, x, dtype=x.dtype)


def square(x, /):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in square")
    return elemwise(nxp.square, x, dtype=x.dtype)


def subtract(x1, x2, /):
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in subtract")
    return elemwise(nxp.subtract, x1, x2, dtype=result_type(x1, x2))


def tan(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in tan")
    return elemwise(nxp.tan, x, dtype=x.dtype)


def tanh(x, /):
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in tanh")
    return elemwise(nxp.tanh, x, dtype=x.dtype)


def trunc(x, /):
    if x.dtype not in _real_numeric_dtypes:
        raise TypeError("Only real numeric dtypes are allowed in trunc")
    if x.dtype in _integer_dtypes:
        # Note: The return dtype of trunc is the same as the input
        return x
    return elemwise(nxp.trunc, x, dtype=x.dtype)



from functools import lru_cache
from typing import Optional, Union

import donfig
from donfig.config_obj import expand_environment_variables

from cubed.runtime.create import create_executor
from cubed.runtime.types import Executor
from cubed.utils import convert_to_bytes


class Spec:
    """Specification of resources available to run a computation."""

    def __init__(
        self,
        work_dir: Union[str, None] = None,
        allowed_mem: Union[int, str, None] = None,
        reserved_mem: Union[int, str, None] = 0,
        executor: Union[Executor, None] = None,
        executor_name: Optional[str] = None,
        executor_options: Optional[dict] = None,
        storage_options: Union[dict, None] = None,
    ):
        """
        Specify resources available to run a computation.

        Parameters
        ----------
        work_dir : str or None
            The directory path (specified as an fsspec URL) used for storing intermediate data.
        allowed_mem : int or str, optional
            The total memory available to a worker for running a task, in bytes.

            If int it should be >=0. If str it should be of form <value><unit> where unit can be kB, MB, GB, TB etc.
            This includes any ``reserved_mem`` that has been set.
        reserved_mem : int or str, optional
            The memory reserved on a worker for non-data use when running a task, in bytes.

            If int it should be >=0. If str it should be of form <value><unit> where unit can be kB, MB, GB, TB etc.
        executor : Executor, optional
            The default executor for running computations.
        storage_options : dict, optional
            Storage options to be passed to fsspec.
        """

        self._work_dir = work_dir

        self._reserved_mem = convert_to_bytes(reserved_mem or 0)
        if allowed_mem is None:
            self._allowed_mem = self.reserved_mem
        else:
            self._allowed_mem = convert_to_bytes(allowed_mem)

        self._executor: Optional[Executor]
        if executor is not None:
            self._executor = executor
        elif executor_name is not None:
            self._executor = create_executor(executor_name, executor_options)
        else:
            self._executor = None

        self._storage_options = storage_options

    @property
    def work_dir(self) -> Optional[str]:
        """The directory path (specified as an fsspec URL) used for storing intermediate data."""
        return self._work_dir

    @property
    def allowed_mem(self) -> int:
        """
        The total memory available to a worker for running a task, in bytes.

        This includes any ``reserved_mem`` that has been set.
        """
        return self._allowed_mem

    @property
    def reserved_mem(self) -> int:
        """
        The memory reserved on a worker for non-data use when running a task, in bytes.

        See Also
        --------
        cubed.measure_reserved_mem
        """
        return self._reserved_mem

    @property
    def executor(self) -> Optional[Executor]:
        """The default executor for running computations."""
        return self._executor

    @property
    def storage_options(self) -> Optional[dict]:
        """Storage options to be passed to fsspec."""
        return self._storage_options

    def __repr__(self) -> str:
        return (
            f"cubed.Spec(work_dir={self._work_dir}, allowed_mem={self._allowed_mem}, "
            f"reserved_mem={self._reserved_mem}, executor={self._executor}, storage_options={self._storage_options})"
        )

    def __eq__(self, other):
        if isinstance(other, Spec):
            return (
                self.work_dir == other.work_dir
                and self.allowed_mem == other.allowed_mem
                and self.reserved_mem == other.reserved_mem
                and self.executor == other.executor
                and self.storage_options == other.storage_options
            )
        else:
            return False


def spec_from_config(config):
    return _spec_from_serialized_config(config.serialize())


@lru_cache  # ensure arrays have the same Spec object for a given config
def _spec_from_serialized_config(ser: str):
    config = donfig.deserialize(ser)
    spec_dict = expand_environment_variables(config["spec"])
    return Spec(**spec_dict)



from numbers import Integral
from typing import Any

import numpy as np
import zarr
from zarr.indexing import BasicIndexer, is_slice

from cubed.backend_array_api import backend_array_to_numpy_array
from cubed.backend_array_api import namespace as nxp
from cubed.backend_array_api import numpy_array_to_backend_array
from cubed.types import T_DType, T_RegularChunks, T_Shape
from cubed.utils import array_memory, broadcast_trick, memory_repr


class VirtualEmptyArray:
    """An array that is never materialized (in memory or on disk) and contains empty values."""

    def __init__(
        self,
        shape: T_Shape,
        dtype: T_DType,
        chunks: T_RegularChunks,
    ):
        # use an empty in-memory Zarr array as a template since it normalizes its properties
        template = zarr.empty(
            shape, dtype=dtype, chunks=chunks, store=zarr.storage.MemoryStore()
        )
        self.shape = template.shape
        self.dtype = template.dtype
        self.chunks = template.chunks
        self.template = template

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)
        indexer = BasicIndexer(key, self.template)
        # use broadcast trick so array chunks only occupy a single value in memory
        return broadcast_trick(nxp.empty)(indexer.shape, dtype=self.dtype)

    @property
    def chunkmem(self):
        # take broadcast trick into account
        return array_memory(self.dtype, (1,))

    @property
    def oindex(self):
        return self.template.oindex


class VirtualFullArray:
    """An array that is never materialized (in memory or on disk) and contains a single fill value."""

    def __init__(
        self,
        shape: T_Shape,
        dtype: T_DType,
        chunks: T_RegularChunks,
        fill_value: Any = None,
    ):
        # use an empty in-memory Zarr array as a template since it normalizes its properties
        template = zarr.full(
            shape,
            fill_value,
            dtype=dtype,
            chunks=chunks,
            store=zarr.storage.MemoryStore(),
        )
        self.shape = template.shape
        self.dtype = template.dtype
        self.chunks = template.chunks
        self.template = template
        self.fill_value = fill_value

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)
        indexer = BasicIndexer(key, self.template)
        # use broadcast trick so array chunks only occupy a single value in memory
        return broadcast_trick(nxp.full)(
            indexer.shape, fill_value=self.fill_value, dtype=self.dtype
        )

    @property
    def chunkmem(self):
        # take broadcast trick into account
        return array_memory(self.dtype, (1,))

    @property
    def oindex(self):
        return self.template.oindex


class VirtualOffsetsArray:
    """An array that is never materialized (in memory or on disk) and contains sequentially incrementing integers."""

    def __init__(self, shape: T_Shape):
        dtype = nxp.int32
        chunks = (1,) * len(shape)
        # use an empty in-memory Zarr array as a template since it normalizes its properties
        template = zarr.empty(
            shape, dtype=dtype, chunks=chunks, store=zarr.storage.MemoryStore()
        )
        self.shape = template.shape
        self.dtype = template.dtype
        self.chunks = template.chunks
        self.ndim = template.ndim

    def __getitem__(self, key):
        if key == () and self.shape == ():
            return nxp.asarray(0, dtype=self.dtype)
        return numpy_array_to_backend_array(
            np.ravel_multi_index(_key_to_index_tuple(key), self.shape), dtype=self.dtype
        )


class VirtualInMemoryArray:
    """A small array that is held in memory but never materialized on disk."""

    def __init__(
        self,
        array: np.ndarray,  # TODO: generalise to array API type
        chunks: T_RegularChunks,
        max_nbytes: int = 10**6,
    ):
        if array.nbytes > max_nbytes:
            raise ValueError(
                f"Size of in memory array is {memory_repr(array.nbytes)} which exceeds maximum of {memory_repr(max_nbytes)}. Consider loading the array from storage using `from_array`."
            )
        self.array = array
        # use an in-memory Zarr array as a template since it normalizes its properties
        # and is needed for oindex
        template = zarr.empty(
            array.shape,
            dtype=array.dtype,
            chunks=chunks,
            store=zarr.storage.MemoryStore(),
        )
        self.shape = template.shape
        self.dtype = template.dtype
        self.chunks = template.chunks
        self.template = template
        if array.size > 0:
            template[...] = backend_array_to_numpy_array(array)

    def __getitem__(self, key):
        return self.array.__getitem__(key)

    @property
    def oindex(self):
        return self.template.oindex


def _key_to_index_tuple(selection):
    if isinstance(selection, slice):
        selection = (selection,)
    assert all(isinstance(s, (slice, Integral)) for s in selection)
    sel = []
    for s in selection:
        if isinstance(s, Integral):
            sel.append(s)
        elif is_slice(s) and s.stop == s.start + 1 and (s.step is None or s.step == 1):
            sel.append(s.start)
        else:
            raise NotImplementedError(f"Offset selection not supported: {selection}")
    return tuple(sel)


def virtual_empty(
    shape: T_Shape, *, dtype: T_DType, chunks: T_RegularChunks, **kwargs
) -> VirtualEmptyArray:
    return VirtualEmptyArray(shape, dtype, chunks, **kwargs)


def virtual_full(
    shape: T_Shape,
    fill_value: Any,
    *,
    dtype: T_DType,
    chunks: T_RegularChunks,
    **kwargs,
) -> VirtualFullArray:
    return VirtualFullArray(shape, dtype, chunks, fill_value, **kwargs)


def virtual_offsets(shape: T_Shape) -> VirtualOffsetsArray:
    return VirtualOffsetsArray(shape)


def virtual_in_memory(
    array: np.ndarray,
    chunks: T_RegularChunks,
) -> VirtualInMemoryArray:
    return VirtualInMemoryArray(array, chunks)



import asyncio
import itertools
from functools import partial

import pytest

from cubed.tests.runtime.utils import check_invocation_counts, deterministic_failure

pytest.importorskip("dask.distributed")

from dask.distributed import Client

from cubed.runtime.executors.dask_distributed_async import map_unordered


async def run_test(function, input, retries, use_backups=False, batch_size=None):
    outputs = set()
    async with Client(asynchronous=True) as client:
        async for output in map_unordered(
            client,
            function,
            input,
            retries=retries,
            use_backups=use_backups,
            batch_size=batch_size,
        ):
            outputs.add(output)
    return outputs


# fmt: off
@pytest.mark.parametrize(
    "timing_map, n_tasks, retries",
    [
        # no failures
        ({}, 3, 2),
        # first invocation fails
        ({0: [-1], 1: [-1], 2: [-1]}, 3, 2),
        # first two invocations fail
        ({0: [-1, -1], 1: [-1, -1], 2: [-1, -1]}, 3, 2),
        # first input sleeps once (not tested since timeout is not supported)
        # ({0: [20]}, 3, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
def test_success(tmp_path, timing_map, n_tasks, retries, use_backups):
    outputs = asyncio.run(
        run_test(
            function=partial(deterministic_failure, tmp_path, timing_map),
            input=range(n_tasks),
            retries=retries,
            use_backups=use_backups,
        )
    )

    assert outputs == set(range(n_tasks))

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    "timing_map, n_tasks, retries",
    [
        # too many failures
        ({0: [-1], 1: [-1], 2: [-1, -1, -1]}, 3, 2),
    ],
)
# fmt: on
@pytest.mark.parametrize("use_backups", [False, True])
def test_failure(tmp_path, timing_map, n_tasks, retries, use_backups):
    with pytest.raises(RuntimeError):
        asyncio.run(
            run_test(
                function=partial(deterministic_failure, tmp_path, timing_map),
                input=range(n_tasks),
                retries=retries,
                use_backups=use_backups,
            )
        )

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    "timing_map, n_tasks, retries",
    [
        ({0: [60]}, 10, 2),
    ],
)
# fmt: on
def test_stragglers(tmp_path, timing_map, n_tasks, retries):
    outputs = asyncio.run(
        run_test(
            function=partial(deterministic_failure, tmp_path, timing_map),
            input=range(n_tasks),
            retries=retries,
            use_backups=True,
        )
    )

    assert outputs == set(range(n_tasks))

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)


def test_batch(tmp_path):
    # input is unbounded, so if entire input were consumed and not read
    # in batches then it would never return, since it would never
    # run the first (failing) input
    with pytest.raises(RuntimeError):
        asyncio.run(
            run_test(
                function=partial(deterministic_failure, tmp_path, {0: [-1]}),
                input=itertools.count(),
                retries=0,
                batch_size=10,
            )
        )



import contextlib
import sys

from toolz import map

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback


class TqdmProgressBar(Callback):
    """Progress bar for a computation."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def on_compute_start(self, event):
        from tqdm.auto import tqdm

        self.pbars = {}
        i = 0
        for name, node in visit_nodes(event.dag, event.resume):
            num_tasks = node["primitive_op"].num_tasks
            op_display_name = node["op_display_name"].replace("\n", " ")
            self.pbars[name] = tqdm(
                *self.args,
                desc=op_display_name,
                total=num_tasks,
                position=i,
                **self.kwargs,
            )
            i = i + 1

    def on_compute_end(self, event):
        for pbar in self.pbars.values():
            pbar.close()

    def on_task_end(self, event):
        self.pbars[event.name].update(event.num_tasks)


@contextlib.contextmanager
def std_out_err_redirect_tqdm():
    """
    Context manager for redirecting stdout and stderr when using tqdm.
    See https://github.com/tqdm/tqdm#redirecting-writing
    """
    from tqdm.contrib import DummyTqdmFile

    orig_out_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = map(DummyTqdmFile, orig_out_err)
        yield orig_out_err[0]
    # Relay exceptions
    except Exception as exc:
        raise exc
    # Always restore sys.stdout/err if necessary
    finally:
        sys.stdout, sys.stderr = orig_out_err



from typing import Optional

from cubed.runtime.types import Executor


def create_executor(name: str, executor_options: Optional[dict] = None) -> Executor:
    """Create an executor from an executor name."""
    executor_options = executor_options or {}
    if name == "beam":
        from cubed.runtime.executors.beam import BeamDagExecutor

        return BeamDagExecutor(**executor_options)
    elif name == "coiled":
        from cubed.runtime.executors.coiled import CoiledFunctionsDagExecutor

        return CoiledFunctionsDagExecutor(**executor_options)
    elif name == "dask":
        from cubed.runtime.executors.dask_distributed_async import (
            AsyncDaskDistributedExecutor,
        )

        return AsyncDaskDistributedExecutor(**executor_options)
    elif name == "lithops":
        from cubed.runtime.executors.lithops import LithopsDagExecutor

        return LithopsDagExecutor(**executor_options)
    elif name == "modal":
        from cubed.runtime.executors.modal_async import AsyncModalDagExecutor

        return AsyncModalDagExecutor(**executor_options)
    elif name == "processes":
        from cubed.runtime.executors.python_async import AsyncPythonDagExecutor

        return AsyncPythonDagExecutor(retries=0, use_processes=True, **executor_options)
    elif name == "single-threaded":
        from cubed.runtime.executors.python import PythonDagExecutor

        return PythonDagExecutor(**executor_options)
    elif name == "threads":
        from cubed.runtime.executors.python_async import AsyncPythonDagExecutor

        return AsyncPythonDagExecutor(**executor_options)
    else:
        raise ValueError(f"Unrecognized executor name: {name}")



from bisect import bisect
from operator import add, mul

import numpy as np
import tlz
from toolz import reduce

from cubed.array_api.creation_functions import empty
from cubed.backend_array_api import namespace as nxp
from cubed.core import squeeze  # noqa: F401
from cubed.core import blockwise, rechunk, unify_chunks
from cubed.core.ops import elemwise, general_blockwise, map_blocks, map_direct
from cubed.utils import block_id_to_offset, get_item, offset_to_block_id, to_chunksize
from cubed.vendor.dask.array.core import broadcast_chunks, normalize_chunks
from cubed.vendor.dask.array.reshape import reshape_rechunk
from cubed.vendor.dask.array.utils import validate_axis


def broadcast_arrays(*arrays):
    # From dask broadcast_arrays

    # Unify uneven chunking
    inds = [list(reversed(range(x.ndim))) for x in arrays]
    uc_args = tlz.concat(zip(arrays, inds))
    _, args = unify_chunks(*uc_args, warn=False)

    shape = np.broadcast_shapes(*(e.shape for e in args))
    chunks = broadcast_chunks(*(e.chunks for e in args))

    result = tuple(broadcast_to(e, shape=shape, chunks=chunks) for e in args)

    return result


def broadcast_to(x, /, shape, *, chunks=None):
    if x.shape == shape and (chunks is None or chunks == x.chunks):
        return x
    ndim_new = len(shape) - x.ndim
    if ndim_new < 0 or any(
        new != old for new, old in zip(shape[ndim_new:], x.shape) if old != 1
    ):
        raise ValueError(f"cannot broadcast shape {x.shape} to shape {shape}")

    # TODO: fix case where shape has a dimension of size zero

    if chunks is None:
        # New dimensions and broadcast dimensions have chunk size 1
        # This behaviour differs from dask where it is the full dimension size
        xchunks = normalize_chunks(x.chunks, x.shape, dtype=x.dtype)
        chunks = tuple((1,) * s for s in shape[:ndim_new]) + tuple(
            bd if old > 1 else ((1,) * new if new > 0 else (0,))
            for bd, old, new in zip(xchunks, x.shape, shape[ndim_new:])
        )
    else:
        chunks = normalize_chunks(
            chunks, shape, dtype=x.dtype, previous_chunks=x.chunks
        )
        for old_bd, new_bd in zip(x.chunks, chunks[ndim_new:]):
            if old_bd != new_bd and old_bd != (1,):
                raise ValueError(
                    "cannot broadcast chunks %s to chunks %s: "
                    "new chunks must either be along a new "
                    "dimension or a dimension of size 1" % (x.chunks, chunks)
                )

    # create an empty array as a template for blockwise to do broadcasting
    template = empty(shape, dtype=nxp.int8, chunks=chunks, spec=x.spec)

    return elemwise(_broadcast_like, x, template, dtype=x.dtype)


def _broadcast_like(x, template):
    return nxp.broadcast_to(x, template.shape)


def concat(arrays, /, *, axis=0):
    if not arrays:
        raise ValueError("Need array(s) to concat")

    if axis is None:
        arrays = [flatten(array) for array in arrays]
        axis = 0

    # TODO: check arrays all have same shape (except in the dimension specified by axis)
    # TODO: type promotion
    # TODO: unify chunks

    a = arrays[0]

    # offsets along axis for the start of each array
    offsets = [0] + list(tlz.accumulate(add, [a.shape[axis] for a in arrays]))

    axis = validate_axis(axis, a.ndim)
    shape = a.shape[:axis] + (offsets[-1],) + a.shape[axis + 1 :]
    dtype = a.dtype
    chunks = normalize_chunks(to_chunksize(a.chunks), shape=shape, dtype=dtype)

    # memory allocated by reading one chunk from input array
    # note that although the output chunk will overlap multiple input chunks,
    # the chunks are read in series, reusing memory
    extra_projected_mem = a.chunkmem

    return map_direct(
        _read_concat_chunk,
        *arrays,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        extra_projected_mem=extra_projected_mem,
        axis=axis,
        offsets=offsets,
    )


def _read_concat_chunk(x, *arrays, axis=None, offsets=None, block_id=None):
    # determine the start and stop indexes for this block along the axis dimension
    chunks = arrays[0].zarray.chunks
    start = block_id[axis] * chunks[axis]
    stop = start + x.shape[axis]

    # produce a key that has slices (except for axis dimension, which is replaced below)
    idx = tuple(0 if i == axis else v for i, v in enumerate(block_id))
    key = get_item(arrays[0].chunks, idx)

    # concatenate slices of the arrays
    parts = []
    for ai, sl in _array_slices(offsets, start, stop):
        key = tuple(sl if i == axis else k for i, k in enumerate(key))
        parts.append(arrays[ai].zarray[key])
    return nxp.concat(parts, axis=axis)


def _array_slices(offsets, start, stop):
    """Return pairs of array index and slice to slice from start to stop in the concatenated array."""
    slice_start = start
    while slice_start < stop:
        # find array that slice_start falls in
        i = bisect(offsets, slice_start) - 1
        slice_stop = min(stop, offsets[i + 1])
        yield i, slice(slice_start - offsets[i], slice_stop - offsets[i])
        slice_start = slice_stop


def expand_dims(x, /, *, axis):
    if not isinstance(axis, tuple):
        axis = (axis,)
    ndim_new = len(axis) + x.ndim
    axis = validate_axis(axis, ndim_new)

    chunks_it = iter(x.chunks)
    chunks = tuple(1 if i in axis else next(chunks_it) for i in range(ndim_new))

    return map_blocks(
        _expand_dims, x, dtype=x.dtype, chunks=chunks, new_axis=axis, axis=axis
    )


def _expand_dims(a, *args, **kwargs):
    if isinstance(a, dict):
        return {k: nxp.expand_dims(v, *args, **kwargs) for k, v in a.items()}
    return nxp.expand_dims(a, *args, **kwargs)


def flatten(x):
    return reshape(x, (-1,))


def moveaxis(
    x,
    source,
    destination,
    /,
):
    # From NumPy: https://github.com/numpy/numpy/blob/a4120979d216cce00dcee511aad70bf7b45ef6e0/numpy/core/numeric.py#L1389-L1457
    from numpy.core.numeric import normalize_axis_tuple

    source = normalize_axis_tuple(source, x.ndim, "source")
    destination = normalize_axis_tuple(destination, x.ndim, "destination")
    if len(source) != len(destination):
        raise ValueError(
            "`source` and `destination` arguments must have "
            "the same number of elements"
        )

    order = [n for n in range(x.ndim) if n not in source]

    for dest, src in sorted(zip(destination, source)):
        order.insert(dest, src)

    result = permute_dims(x, order)
    return result


def permute_dims(x, /, axes):
    # From dask transpose
    if axes:
        if len(axes) != x.ndim:
            raise ValueError("axes don't match array")
    else:
        axes = tuple(range(x.ndim))[::-1]
    axes = tuple(d + x.ndim if d < 0 else d for d in axes)

    # extra memory copy due to Zarr enforcing C order on transposed array
    extra_projected_mem = x.chunkmem
    return blockwise(
        nxp.permute_dims,
        axes,
        x,
        tuple(range(x.ndim)),
        dtype=x.dtype,
        axes=axes,
        extra_projected_mem=extra_projected_mem,
    )


def reshape(x, /, shape, *, copy=None):
    # based on dask reshape

    known_sizes = [s for s in shape if s != -1]
    if len(known_sizes) != len(shape):
        if len(shape) - len(known_sizes) > 1:
            raise ValueError("can only specify one unknown dimension")
        # Fastpath for x.reshape(-1) on 1D arrays
        if len(shape) == 1 and x.ndim == 1:
            return x
        missing_size = x.size // reduce(mul, known_sizes, 1)
        shape = tuple(missing_size if s == -1 else s for s in shape)

    if reduce(mul, shape, 1) != x.size:
        raise ValueError("total size of new array must be unchanged")

    if x.shape == shape:
        return x

    if x.npartitions == 1:
        outchunks = tuple((d,) for d in shape)
        return reshape_chunks(x, shape, outchunks)

    inchunks, outchunks = reshape_rechunk(x.shape, shape, x.chunks)

    # TODO: make sure chunks are not too large

    x2 = rechunk(x, to_chunksize(inchunks))

    return reshape_chunks(x2, shape, outchunks)


def reshape_chunks(x, shape, chunks):
    if reduce(mul, shape, 1) != x.size:
        raise ValueError("total size of new array must be unchanged")

    # TODO: check number of chunks is unchanged
    # inchunks = normalize_chunks(x.chunks, shape=x.shape, dtype=x.dtype)
    outchunks = normalize_chunks(chunks, shape=shape, dtype=x.dtype)

    # use an empty template (handles smaller end chunks)
    template = empty(shape, dtype=x.dtype, chunks=chunks, spec=x.spec)

    def key_function(out_key):
        out_coords = out_key[1:]
        offset = block_id_to_offset(out_coords, template.numblocks)
        in_coords = offset_to_block_id(offset, x.numblocks)
        return (
            (x.name, *in_coords),
            (template.name, *out_coords),
        )

    return general_blockwise(
        _reshape_chunk,
        key_function,
        x,
        template,
        shape=shape,
        dtype=x.dtype,
        chunks=outchunks,
    )


def _reshape_chunk(x, template):
    return nxp.reshape(x, template.shape)


def stack(arrays, /, *, axis=0):
    if not arrays:
        raise ValueError("Need array(s) to stack")

    # TODO: check arrays all have same shape
    # TODO: type promotion
    # TODO: unify chunks

    a = arrays[0]

    axis = validate_axis(axis, a.ndim + 1)
    shape = a.shape[:axis] + (len(arrays),) + a.shape[axis:]
    dtype = a.dtype
    chunks = a.chunks[:axis] + ((1,) * len(arrays),) + a.chunks[axis:]

    array_names = [a.name for a in arrays]

    def key_function(out_key):
        out_coords = out_key[1:]
        in_name = array_names[out_coords[axis]]
        return ((in_name, *(out_coords[:axis] + out_coords[(axis + 1) :])),)

    # We have to mark this as fusable=False since the number of input args to
    # the _read_stack_chunk function is *not* the same as the number of
    # predecessor nodes in the DAG, and the fusion functions in blockwise
    # assume they are the same. See https://github.com/cubed-dev/cubed/issues/414
    return general_blockwise(
        _read_stack_chunk,
        key_function,
        *arrays,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        axis=axis,
        fusable=False,
    )


def _read_stack_chunk(array, axis=None):
    return nxp.expand_dims(array, axis=axis)



import math
from functools import reduce
from operator import mul

_not_implemented_message = """
Dask's reshape only supports operations that merge or split existing dimensions
evenly. For example:

>>> x = da.ones((6, 5, 4), chunks=(3, 2, 2))
>>> x.reshape((3, 2, 5, 4))  # supported, splits 6 into 3 & 2
>>> x.reshape((30, 4))       # supported, merges 6 & 5 into 30
>>> x.reshape((4, 5, 6))     # unsupported, existing dimensions split unevenly

To work around this you may call reshape in multiple passes, or (if your data
is small enough) call ``compute`` first and handle reshaping in ``numpy``
directly.
"""


def reshape_rechunk(inshape, outshape, inchunks):
    assert all(isinstance(c, tuple) for c in inchunks)
    ii = len(inshape) - 1
    oi = len(outshape) - 1
    result_inchunks = [None for i in range(len(inshape))]
    result_outchunks = [None for i in range(len(outshape))]

    while ii >= 0 or oi >= 0:
        if inshape[ii] == outshape[oi]:
            result_inchunks[ii] = inchunks[ii]
            result_outchunks[oi] = inchunks[ii]
            ii -= 1
            oi -= 1
            continue
        din = inshape[ii]
        dout = outshape[oi]
        if din == 1:
            result_inchunks[ii] = (1,)
            ii -= 1
        elif dout == 1:
            result_outchunks[oi] = (1,)
            oi -= 1
        elif din < dout:  # (4, 4, 4) -> (64,)
            ileft = ii - 1
            while (
                ileft >= 0 and reduce(mul, inshape[ileft : ii + 1]) < dout
            ):  # 4 < 64, 4*4 < 64, 4*4*4 == 64
                ileft -= 1
            if reduce(mul, inshape[ileft : ii + 1]) != dout:
                raise NotImplementedError(_not_implemented_message)
            # Special case to avoid intermediate rechunking:
            # When all the lower axis are completely chunked (chunksize=1) then
            # we're simply moving around blocks.
            if all(len(inchunks[i]) == inshape[i] for i in range(ii)):
                for i in range(ii + 1):
                    result_inchunks[i] = inchunks[i]
                result_outchunks[oi] = inchunks[ii] * math.prod(
                    map(len, inchunks[ileft:ii])
                )
            else:
                for i in range(ileft + 1, ii + 1):  # need single-shape dimensions
                    result_inchunks[i] = (inshape[i],)  # chunks[i] = (4,)

                chunk_reduction = reduce(mul, map(len, inchunks[ileft + 1 : ii + 1]))
                result_inchunks[ileft] = expand_tuple(inchunks[ileft], chunk_reduction)

                prod = reduce(mul, inshape[ileft + 1 : ii + 1])  # 16
                result_outchunks[oi] = tuple(
                    prod * c for c in result_inchunks[ileft]
                )  # (1, 1, 1, 1) .* 16

            oi -= 1
            ii = ileft - 1
        elif din > dout:  # (64,) -> (4, 4, 4)
            oleft = oi - 1
            while oleft >= 0 and reduce(mul, outshape[oleft : oi + 1]) < din:
                oleft -= 1
            if reduce(mul, outshape[oleft : oi + 1]) != din:
                raise NotImplementedError(_not_implemented_message)
            # TODO: don't coalesce shapes unnecessarily
            cs = reduce(mul, outshape[oleft + 1 : oi + 1])

            result_inchunks[ii] = contract_tuple(inchunks[ii], cs)  # (16, 16, 16, 16)

            for i in range(oleft + 1, oi + 1):
                result_outchunks[i] = (outshape[i],)

            result_outchunks[oleft] = tuple(c // cs for c in result_inchunks[ii])

            oi = oleft - 1
            ii -= 1

    return tuple(result_inchunks), tuple(result_outchunks)


def expand_tuple(chunks, factor):
    """

    >>> expand_tuple((2, 4), 2)
    (1, 1, 2, 2)

    >>> expand_tuple((2, 4), 3)
    (1, 1, 1, 1, 2)

    >>> expand_tuple((3, 4), 2)
    (1, 2, 2, 2)

    >>> expand_tuple((7, 4), 3)
    (2, 2, 3, 1, 1, 2)
    """
    if factor == 1:
        return chunks

    out = []
    for c in chunks:
        x = c
        part = max(x / factor, 1)
        while x >= 2 * part:
            out.append(int(part))
            x -= int(part)
        if x:
            out.append(x)
    assert sum(chunks) == sum(out)
    return tuple(out)


def contract_tuple(chunks, factor):
    """Return simple chunks tuple such that factor divides all elements

    Examples
    --------

    >>> contract_tuple((2, 2, 8, 4), 4)
    (4, 8, 4)
    """
    assert sum(chunks) % factor == 0

    out = []
    residual = 0
    for chunk in chunks:
        chunk += residual
        div = chunk // factor
        residual = chunk % factor
        good = factor * div
        if good:
            out.append(good)
    return tuple(out)



from __future__ import annotations

import functools
import inspect
import re
from datetime import datetime
from numbers import Integral
from typing import Any


def is_integer(i) -> bool:
    """
    >>> is_integer(6)
    True
    >>> is_integer(42.0)
    True
    >>> is_integer('abc')
    False
    """
    return isinstance(i, Integral) or (isinstance(i, float) and i.is_integer())


def typename(typ: Any, short: bool = False) -> str:
    """
    Return the name of a type

    Examples
    --------
    >>> typename(int)
    'int'

    >>> from dask.core import literal
    >>> typename(literal)
    'dask.core.literal'
    >>> typename(literal, short=True)
    'dask.literal'
    """
    if not isinstance(typ, type):
        return typename(type(typ))
    try:
        if not typ.__module__ or typ.__module__ == "builtins":
            return typ.__name__
        else:
            if short:
                module, *_ = typ.__module__.split(".")
            else:
                module = typ.__module__
            return module + "." + typ.__name__
    except AttributeError:
        return str(typ)


def parse_bytes(s: float | str) -> int:
    """Parse byte string to numbers

    >>> from dask.utils import parse_bytes
    >>> parse_bytes('100')
    100
    >>> parse_bytes('100 MB')
    100000000
    >>> parse_bytes('100M')
    100000000
    >>> parse_bytes('5kB')
    5000
    >>> parse_bytes('5.4 kB')
    5400
    >>> parse_bytes('1kiB')
    1024
    >>> parse_bytes('1e6')
    1000000
    >>> parse_bytes('1e6 kB')
    1000000000
    >>> parse_bytes('MB')
    1000000
    >>> parse_bytes(123)
    123
    >>> parse_bytes('5 foos')
    Traceback (most recent call last):
        ...
    ValueError: Could not interpret 'foos' as a byte unit
    """
    if isinstance(s, (int, float)):
        return int(s)
    s = s.replace(" ", "")
    if not any(char.isdigit() for char in s):
        s = "1" + s

    for i in range(len(s) - 1, -1, -1):
        if not s[i].isalpha():
            break
    index = i + 1

    prefix = s[:index]
    suffix = s[index:]

    try:
        n = float(prefix)
    except ValueError as e:
        raise ValueError("Could not interpret '%s' as a number" % prefix) from e

    try:
        multiplier = byte_sizes[suffix.lower()]
    except KeyError as e:
        raise ValueError("Could not interpret '%s' as a byte unit" % suffix) from e

    result = n * multiplier
    return int(result)


byte_sizes = {
    "kB": 10**3,
    "MB": 10**6,
    "GB": 10**9,
    "TB": 10**12,
    "PB": 10**15,
    "KiB": 2**10,
    "MiB": 2**20,
    "GiB": 2**30,
    "TiB": 2**40,
    "PiB": 2**50,
    "B": 1,
    "": 1,
}
byte_sizes = {k.lower(): v for k, v in byte_sizes.items()}
byte_sizes.update({k[0]: v for k, v in byte_sizes.items() if k and "i" not in k})
byte_sizes.update({k[:-1]: v for k, v in byte_sizes.items() if k and "i" in k})


def has_keyword(func, keyword):
    try:
        return keyword in inspect.signature(func).parameters
    except Exception:
        return False


def format_time(n: float) -> str:
    """format integers as time

    >>> from dask.utils import format_time
    >>> format_time(1)
    '1.00 s'
    >>> format_time(0.001234)
    '1.23 ms'
    >>> format_time(0.00012345)
    '123.45 us'
    >>> format_time(123.456)
    '123.46 s'
    >>> format_time(1234.567)
    '20m 34s'
    >>> format_time(12345.67)
    '3hr 25m'
    >>> format_time(123456.78)
    '34hr 17m'
    >>> format_time(1234567.89)
    '14d 6hr'
    """
    if n > 24 * 60 * 60 * 2:
        d = int(n / 3600 / 24)
        h = int((n - d * 3600 * 24) / 3600)
        return f"{d}d {h}hr"
    if n > 60 * 60 * 2:
        h = int(n / 3600)
        m = int((n - h * 3600) / 60)
        return f"{h}hr {m}m"
    if n > 60 * 10:
        m = int(n / 60)
        s = int(n - m * 60)
        return f"{m}m {s}s"
    if n >= 1:
        return "%.2f s" % n
    if n >= 1e-3:
        return "%.2f ms" % (n * 1e3)
    return "%.2f us" % (n * 1e6)


def format_time_ago(n: datetime) -> str:
    """Calculate a '3 hours ago' type string from a Python datetime.

    Examples
    --------
    >>> from datetime import datetime, timedelta

    >>> now = datetime.now()
    >>> format_time_ago(now)
    'Just now'

    >>> past = datetime.now() - timedelta(minutes=1)
    >>> format_time_ago(past)
    '1 minute ago'

    >>> past = datetime.now() - timedelta(minutes=2)
    >>> format_time_ago(past)
    '2 minutes ago'

    >>> past = datetime.now() - timedelta(hours=1)
    >>> format_time_ago(past)
    '1 hour ago'

    >>> past = datetime.now() - timedelta(hours=6)
    >>> format_time_ago(past)
    '6 hours ago'

    >>> past = datetime.now() - timedelta(days=1)
    >>> format_time_ago(past)
    '1 day ago'

    >>> past = datetime.now() - timedelta(days=5)
    >>> format_time_ago(past)
    '5 days ago'

    >>> past = datetime.now() - timedelta(days=8)
    >>> format_time_ago(past)
    '1 week ago'

    >>> past = datetime.now() - timedelta(days=16)
    >>> format_time_ago(past)
    '2 weeks ago'

    >>> past = datetime.now() - timedelta(days=190)
    >>> format_time_ago(past)
    '6 months ago'

    >>> past = datetime.now() - timedelta(days=800)
    >>> format_time_ago(past)
    '2 years ago'

    """
    units = {
        "years": lambda diff: diff.days / 365,
        "months": lambda diff: diff.days / 30.436875,  # Average days per month
        "weeks": lambda diff: diff.days / 7,
        "days": lambda diff: diff.days,
        "hours": lambda diff: diff.seconds / 3600,
        "minutes": lambda diff: diff.seconds % 3600 / 60,
    }
    diff = datetime.now() - n
    for unit in units:
        dur = int(units[unit](diff))
        if dur > 0:
            if dur == 1:  # De-pluralize
                unit = unit[:-1]
            return f"{dur} {unit} ago"
    return "Just now"


def format_bytes(n: int) -> str:
    """Format bytes as text

    >>> from dask.utils import format_bytes
    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.21 kiB'
    >>> format_bytes(12345678)
    '11.77 MiB'
    >>> format_bytes(1234567890)
    '1.15 GiB'
    >>> format_bytes(1234567890000)
    '1.12 TiB'
    >>> format_bytes(1234567890000000)
    '1.10 PiB'

    For all values < 2**60, the output is always <= 10 characters.
    """
    for prefix, k in (
        ("Pi", 2**50),
        ("Ti", 2**40),
        ("Gi", 2**30),
        ("Mi", 2**20),
        ("ki", 2**10),
    ):
        if n >= k * 0.9:
            return f"{n / k:.2f} {prefix}B"
    return f"{n} B"


hex_pattern = re.compile("[a-f]+")


@functools.lru_cache(100000)
def key_split(s):
    """
    >>> key_split('x')
    'x'
    >>> key_split('x-1')
    'x'
    >>> key_split('x-1-2-3')
    'x'
    >>> key_split(('x-2', 1))
    'x'
    >>> key_split("('x-2', 1)")
    'x'
    >>> key_split("('x', 1)")
    'x'
    >>> key_split('hello-world-1')
    'hello-world'
    >>> key_split(b'hello-world-1')
    'hello-world'
    >>> key_split('ae05086432ca935f6eba409a8ecd4896')
    'data'
    >>> key_split('<module.submodule.myclass object at 0xdaf372')
    'myclass'
    >>> key_split(None)
    'Other'
    >>> key_split('x-abcdefab')  # ignores hex
    'x'
    >>> key_split('_(x)')  # strips unpleasant characters
    'x'
    """
    if type(s) is bytes:
        s = s.decode()
    if type(s) is tuple:
        s = s[0]
    try:
        words = s.split("-")
        if not words[0][0].isalpha():
            result = words[0].split(",")[0].strip("_'()\"")
        else:
            result = words[0]
        for word in words[1:]:
            if word.isalpha() and not (
                len(word) == 8 and hex_pattern.match(word) is not None
            ):
                result += "-" + word
            else:
                break
        if len(result) == 32 and re.match(r"[a-f0-9]{32}", result):
            return "data"
        else:
            if result[0] == "<":
                result = result.strip("<>").split()[0].split(".")[-1]
            return result
    except Exception:
        return "Other"



from numbers import Integral
from typing import Iterable

from cubed.array_api.data_type_functions import result_type
from cubed.array_api.dtypes import _numeric_dtypes
from cubed.array_api.manipulation_functions import expand_dims
from cubed.backend_array_api import namespace as nxp
from cubed.core import blockwise, reduction, squeeze


def matmul(x1, x2, /, use_new_impl=True, split_every=None):
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in matmul")

    if x1.ndim == 0 or x2.ndim == 0:
        raise ValueError("matmul does not support 0-dimensional arrays.")

    x1_is_1d = False
    if x1.ndim == 1:
        x1_is_1d = True
        x1 = expand_dims(x1, axis=0)

    x2_is_1d = False
    if x2.ndim == 1:
        x2_is_1d = True
        x2 = expand_dims(x2, axis=-1)

    if x1.ndim < x2.ndim:
        x1 = expand_dims(x1, tuple(range(x2.ndim - x1.ndim)))
    elif x1.ndim > x2.ndim:
        x2 = expand_dims(x2, tuple(range(x1.ndim - x2.ndim)))

    out_ind = tuple(range(x1.ndim + 1))
    x1_ind = tuple(range(x1.ndim))
    x2_ind = tuple(range(x1.ndim - 2)) + (x1_ind[-1], x1.ndim)

    dtype = result_type(x1, x2)

    out = blockwise(
        _matmul,
        out_ind,
        x1,
        x1_ind,
        x2,
        x2_ind,
        adjust_chunks={x1_ind[-1]: 1},
        dtype=dtype,
    )

    out = _sum_wo_cat(
        out, axis=-2, dtype=dtype, use_new_impl=use_new_impl, split_every=split_every
    )

    if x1_is_1d:
        out = squeeze(out, -2)
    if x2_is_1d:
        out = squeeze(out, -1)

    return out


def _matmul(a, b):
    chunk = nxp.matmul(a, b)
    return chunk[..., nxp.newaxis, :]


def _sum_wo_cat(a, axis=None, dtype=None, use_new_impl=True, split_every=None):
    if a.shape[axis] == 1:
        return squeeze(a, axis)

    extra_func_kwargs = dict(dtype=dtype)
    return reduction(
        a,
        _chunk_sum,
        axis=axis,
        dtype=dtype,
        use_new_impl=use_new_impl,
        split_every=split_every,
        extra_func_kwargs=extra_func_kwargs,
    )


def _chunk_sum(a, axis=None, dtype=None, keepdims=None):
    return nxp.sum(a, axis=axis, dtype=dtype, keepdims=True)


def matrix_transpose(x, /):
    if x.ndim < 2:
        raise ValueError("x must be at least 2-dimensional for matrix_transpose")

    from cubed.array_api.manipulation_functions import permute_dims

    axes = list(range(x.ndim))
    axes[-1], axes[-2] = axes[-2], axes[-1]  # swap last two axes
    return permute_dims(x, axes)


def outer(x1, x2, /):
    return blockwise(nxp.linalg.outer, "ij", x1, "i", x2, "j", dtype=x1.dtype)


def tensordot(x1, x2, /, *, axes=2, use_new_impl=True, split_every=None):
    from cubed.array_api.statistical_functions import sum

    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in tensordot")

    # based on dask

    if isinstance(axes, Iterable):
        x1_axes, x2_axes = axes
    else:
        x1_axes = tuple(range(x1.ndim - axes, x1.ndim))
        x2_axes = tuple(range(0, axes))
    if isinstance(x1_axes, Integral):
        x1_axes = (x1_axes,)
    if isinstance(x2_axes, Integral):
        x2_axes = (x2_axes,)
    if isinstance(x1_axes, list):
        x1_axes = tuple(x1_axes)
    if isinstance(x2_axes, list):
        x2_axes = tuple(x2_axes)

    dtype = result_type(x1, x2)

    x1_ind = list(range(x1.ndim))
    x2_ind = list(range(x1.ndim, x1.ndim + x2.ndim))
    out_ind = x1_ind + x2_ind
    adjust_chunks = {}
    for a1, a2 in zip(x1_axes, x2_axes):
        out_ind.remove(x2_ind[a2])
        x2_ind[a2] = x1_ind[a1]
        adjust_chunks[x1_ind[a1]] = lambda c: 1

    out = blockwise(
        _tensordot,
        out_ind,
        x1,
        x1_ind,
        x2,
        x2_ind,
        dtype=dtype,
        adjust_chunks=adjust_chunks,
        axes=(x1_axes, x2_axes),
    )
    return sum(
        out,
        axis=x1_axes,
        dtype=dtype,
        use_new_impl=use_new_impl,
        split_every=split_every,
    )


def _tensordot(a, b, axes):
    x = nxp.tensordot(a, b, axes=axes)
    ind = [slice(None, None)] * x.ndim
    for a in sorted(axes[0]):
        ind.insert(a, None)
    x = x[tuple(ind)]
    return x


def vecdot(x1, x2, /, *, axis=-1, use_new_impl=True, split_every=None):
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in vecdot")
    return tensordot(
        x1,
        x2,
        axes=((axis,), (axis,)),
        use_new_impl=use_new_impl,
        split_every=split_every,
    )



#!/usr/bin/env python
from setuptools import setup

if __name__ == "__main__":
    setup(
        name="cubed",
        package_dir={
            "cubed": "cubed"
        },  # avoids pip install problems if one creates a tmp directory alongside cubed directory
    )



import pytest

pytest.importorskip("lithops")

from lithops.executors import LocalhostExecutor

from cubed.runtime.executors.lithops_retries import RetryingFunctionExecutor
from cubed.tests.runtime.utils import check_invocation_counts, deterministic_failure


def run_test(function, input, retries, timeout=10):
    with RetryingFunctionExecutor(LocalhostExecutor()) as executor:
        futures = executor.map(
            function,
            input,
            timeout=timeout,
            retries=retries,
        )
        done, pending = executor.wait(futures, throw_except=False)
        assert len(pending) == 0
    outputs = set(f.result() for f in done)
    return outputs


# fmt: off
@pytest.mark.parametrize(
    "timing_map, n_tasks, retries",
    [
        # no failures
        ({}, 3, 2),
        # first invocation fails
        ({0: [-1], 1: [-1], 2: [-1]}, 3, 2),
        # first two invocations fail
        ({0: [-1, -1], 1: [-1, -1], 2: [-1, -1]}, 3, 2),
        # first input sleeps once
        ({0: [20]}, 3, 2),
    ],
)
# fmt: on
def test_success(tmp_path, timing_map, n_tasks, retries):
    partial_map_function = lambda x: deterministic_failure(tmp_path, timing_map, x)
    outputs = run_test(
        function=partial_map_function,
        input=range(n_tasks),
        retries=retries,
    )

    assert outputs == set(range(n_tasks))

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)


# fmt: off
@pytest.mark.parametrize(
    "timing_map, n_tasks, retries",
    [
        # too many failures
        ({0: [-1], 1: [-1], 2: [-1, -1, -1]}, 3, 2),
    ],
)
# fmt: on
def test_failure(tmp_path, timing_map, n_tasks, retries):
    partial_map_function = lambda x: deterministic_failure(tmp_path, timing_map, x)
    with pytest.raises(RuntimeError):
        run_test(
            function=partial_map_function,
            input=range(n_tasks),
            retries=retries,
        )

    check_invocation_counts(tmp_path, timing_map, n_tasks, retries)



# Copied from numpy.array_api
from cubed.backend_array_api import namespace as nxp

int8 = nxp.int8
int16 = nxp.int16
int32 = nxp.int32
int64 = nxp.int64
uint8 = nxp.uint8
uint16 = nxp.uint16
uint32 = nxp.uint32
uint64 = nxp.uint64
float32 = nxp.float32
float64 = nxp.float64
complex64 = nxp.complex64
complex128 = nxp.complex128
bool = nxp.bool

_all_dtypes = (
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
    float32,
    float64,
    complex64,
    complex128,
    bool,
)
_boolean_dtypes = (bool,)
_real_floating_dtypes = (float32, float64)
_floating_dtypes = (float32, float64, complex64, complex128)
_complex_floating_dtypes = (complex64, complex128)
_integer_dtypes = (int8, int16, int32, int64, uint8, uint16, uint32, uint64)
_signed_integer_dtypes = (int8, int16, int32, int64)
_unsigned_integer_dtypes = (uint8, uint16, uint32, uint64)
_integer_or_boolean_dtypes = (
    bool,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)
_real_numeric_dtypes = (
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)
_numeric_dtypes = (
    float32,
    float64,
    complex64,
    complex128,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)

_dtype_categories = {
    "all": _all_dtypes,
    "real numeric": _real_numeric_dtypes,
    "numeric": _numeric_dtypes,
    "integer": _integer_dtypes,
    "integer or boolean": _integer_or_boolean_dtypes,
    "boolean": _boolean_dtypes,
    "real floating-point": _floating_dtypes,
    "complex floating-point": _complex_floating_dtypes,
    "floating-point": _floating_dtypes,
}



import numpy as np
import pytest
from numpy.testing import assert_array_equal

import cubed
import cubed.array_api as xp


@pytest.fixture()
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


def test_nanmean(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, xp.nan]], chunks=(2, 2), spec=spec)
    b = cubed.nanmean(a)
    assert_array_equal(
        b.compute(), np.nanmean(np.array([[1, 2, 3], [4, 5, 6], [7, 8, np.nan]]))
    )


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_nanmean_allnan(spec):
    a = xp.asarray([xp.nan], spec=spec)
    b = cubed.nanmean(a)
    assert_array_equal(b.compute(), np.nanmean(np.array([np.nan])))


def test_nansum(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, xp.nan]], chunks=(2, 2), spec=spec)
    b = cubed.nansum(a)
    assert_array_equal(
        b.compute(), np.nansum(np.array([[1, 2, 3], [4, 5, 6], [7, 8, np.nan]]))
    )


def test_nansum_allnan(spec):
    a = xp.asarray([xp.nan], spec=spec)
    b = cubed.nansum(a)
    assert_array_equal(b.compute(), np.nansum(np.array([np.nan])))



import collections
import itertools
import platform
import sys
import sysconfig
import traceback
from collections.abc import Iterator
from dataclasses import dataclass
from functools import partial
from itertools import islice
from math import prod
from operator import add
from pathlib import Path
from posixpath import join
from typing import Dict, Tuple, Union
from urllib.parse import quote, unquote, urlsplit, urlunsplit

import numpy as np
import tlz as toolz

from cubed.backend_array_api import namespace as nxp
from cubed.types import T_DType, T_RectangularChunks, T_RegularChunks, T_Shape
from cubed.vendor.dask.array.core import _check_regular_chunks, normalize_chunks

PathType = Union[str, Path]


def array_memory(dtype: T_DType, shape: T_Shape) -> int:
    """Calculate the amount of memory in bytes that an array uses."""
    return np.dtype(dtype).itemsize * prod(shape)


def chunk_memory(arr) -> int:
    """Calculate the amount of memory in bytes that a single chunk uses."""
    if hasattr(arr, "chunkmem"):
        return arr.chunkmem
    return array_memory(
        arr.dtype,
        to_chunksize(normalize_chunks(arr.chunks, shape=arr.shape, dtype=arr.dtype)),
    )


def offset_to_block_id(offset: int, numblocks: Tuple[int, ...]) -> Tuple[int, ...]:
    """Convert an index offset to a block ID (chunk coordinates)."""
    return tuple(int(i) for i in np.unravel_index(offset, numblocks))


def block_id_to_offset(block_id: Tuple[int, ...], numblocks: Tuple[int, ...]) -> int:
    """Convert a block ID (chunk coordinates) to an index offset."""
    return int(np.ravel_multi_index(block_id, numblocks))


def get_item(chunks: T_RectangularChunks, idx: Tuple[int, ...]) -> Tuple[slice, ...]:
    """Convert a chunk index to a tuple of slices."""
    # could use Dask's cached_cumsum here if it improves performance
    starts = tuple(_cumsum(c, initial_zero=True) for c in chunks)
    loc = tuple((start[i], start[i + 1]) for i, start in zip(idx, starts))
    return tuple(slice(*s, None) for s in loc)


def _cumsum(seq, initial_zero=False):
    if initial_zero:
        return tuple(toolz.accumulate(add, seq, 0))
    else:
        return tuple(toolz.accumulate(add, seq))


def join_path(dir_url: PathType, child_path: str) -> str:
    """Combine a URL for a directory with a child path"""
    parts = urlsplit(str(dir_url))
    new_path = quote(join(unquote(parts.path), child_path))
    split_parts = (parts.scheme, parts.netloc, new_path, parts.query, parts.fragment)
    return urlunsplit(split_parts)


def memory_repr(num: int) -> str:
    """Convert bytes to a human-readable string in decimal form.
    1 KB is 1,000 bytes, 1 MB is 1,000,000 bytes, and so on.

    Parameters
    ----------
    num: int
        Number of bytes

    Returns
    -------
    str
    """
    if num < 0:
        raise ValueError(f"Invalid value: {num}. Expected a positive integer.")
    if num < 1000.0:
        return f"{num} bytes"
    val = num / 1000.0
    for x in ["KB", "MB", "GB", "TB", "PB"]:
        if val < 1000.0:
            return f"{val:3.1f} {x}"
        val /= 1000.0
    # fall back to scientific notation
    return f"{num:.1e} bytes"


def peak_measured_mem() -> int:
    """Measures the peak memory usage in bytes.

    Note: this function currently doesn't work on Windows.
    """

    if platform.system() == "Windows":
        raise NotImplementedError("`peak_measured_mem` is not implemented on Windows")

    from resource import RUSAGE_SELF, getrusage

    ru_maxrss = getrusage(RUSAGE_SELF).ru_maxrss
    # note that on Linux ru_maxrss is in KiB, while on Mac it is in bytes
    # see https://pythonspeed.com/articles/estimating-memory-usage/#measuring-peak-memory-usage
    return ru_maxrss * 1024 if platform.system() == "Linux" else ru_maxrss


def to_chunksize(chunkset: T_RectangularChunks) -> T_RegularChunks:
    """Convert a chunkset to a chunk size for Zarr.

    Parameters
    ----------
    chunkset: tuple of tuples of ints
        From the ``.chunks`` attribute of an ``Array``

    Returns
    -------
    Tuple of ints
    """

    if not _check_regular_chunks(chunkset):
        raise ValueError(f"Array must have regular chunks, but found chunks={chunkset}")

    return tuple(max(c[0], 1) for c in chunkset)


@dataclass
class StackSummary:
    """Like Python's ``FrameSummary``, but with module information."""

    filename: str
    lineno: int
    name: str
    module: str
    array_names_to_variable_names: dict

    def is_cubed(self):
        """Return True if this stack frame is a Cubed call."""
        return self.module.startswith("cubed.") and not self.module.startswith(
            "cubed.tests."
        )

    def is_on_python_lib_path(self):
        """Return True if this stack frame is from a library on Python's library path."""
        python_lib_path = sysconfig.get_path("purelib")

        return self.filename.startswith(python_lib_path)


def extract_stack_summaries(frame, limit=None):
    """Like Python's ``StackSummary.extract``, but returns module information."""
    frame_gen = traceback.walk_stack(frame)

    # from StackSummary.extract
    if limit is None:
        limit = getattr(sys, "tracebacklimit", None)
        if limit is not None and limit < 0:
            limit = 0
    if limit is not None:
        if limit >= 0:
            frame_gen = itertools.islice(frame_gen, limit)
        else:
            frame_gen = collections.deque(frame_gen, maxlen=-limit)
    # end from StackSummary.extract

    stack_summaries = []
    for f, _ in frame_gen:
        module = f.f_globals.get("__name__", "")
        summary = StackSummary(
            filename=f.f_code.co_filename,
            lineno=f.f_lineno,
            name=f.f_code.co_name,
            module=module,
            array_names_to_variable_names=extract_array_names(f),
        )
        stack_summaries.append(summary)
    stack_summaries.reverse()

    return stack_summaries


def extract_array_names(frame):
    """Look for Cubed arrays in local variables to create a mapping from (internally generated) array names to variable names."""

    from cubed import Array

    array_names_to_variable_names = {}
    for var, arr in frame.f_locals.items():
        if isinstance(arr, Array):
            array_names_to_variable_names[arr.name] = var
        elif (
            type(arr).__module__.split(".")[0] == "xarray"
            and arr.__class__.__name__ == "DataArray"
        ):
            if isinstance(arr.data, Array):
                array_names_to_variable_names[arr.data.name] = arr.name
    return array_names_to_variable_names


def convert_to_bytes(size: Union[int, float, str]) -> int:
    """
    Converts the input data size to bytes.

    The data size can be expressed as an integer or as a string with different SI prefixes such as '500kB', '2MB', or '1GB'.

    Parameters
    ----------
    size: int, float, or str:
        Size of data. If numeric it should represent an integer >=0. If str it should be of form <value><unit> where unit can be B, kB, MB, GB, TB etc.

    Returns
    -------
    int: The size in bytes
    """
    units: Dict[str, int] = {"kB": 1, "MB": 2, "GB": 3, "TB": 4, "PB": 5}

    def is_numeric_str(s: str) -> bool:
        try:
            float(s)
            return True
        except ValueError:
            return False

    if isinstance(size, str):
        size = size.replace(" ", "")

        # check if the format of the string is valid
        if is_numeric_str(size):
            unit_factor = 1.0
            value = size
        elif size[-1] == "B" and is_numeric_str(size[:-1]):
            unit_factor = 1.0
            value = size[:-1]
        elif size[-2:] in units and is_numeric_str(size[:-2]):
            unit = size[-2:]
            unit_factor = 1000 ** units[unit]
            value = size[:-2]
        else:
            raise ValueError(
                f"Invalid value: {size}. Expected the string to be a numeric value ending with an SI prefix."
            )

        # convert to float number of bytes
        size = float(value) * unit_factor

    if isinstance(size, float):
        if size.is_integer():
            size = int(size)
        else:
            raise ValueError(
                f"Invalid value: {size}. Can't have a non-integer number of bytes"
            )

    if size >= 0:
        return size
    else:
        raise ValueError(f"Invalid value: {size}. Must be a positive value")


# Based on more_itertools
def split_into(iterable, sizes):
    """Yield a list of sequential items from *iterable* of length 'n' for each
    integer 'n' in *sizes*."""
    it = iter(iterable)
    for size in sizes:
        yield list(islice(it, size))


def map_nested(func, seq):
    """Apply a function inside nested lists or iterators, while preserving
    the nesting, and the collection or iterator type.

    Examples
    --------

    >>> from cubed.utils import map_nested
    >>> inc = lambda x: x + 1
    >>> map_nested(inc, [[1, 2], [3, 4]])
    [[2, 3], [4, 5]]

    >>> it = map_nested(inc, iter([1, 2]))
    >>> next(it)
    2
    >>> next(it)
    3
    """
    if isinstance(seq, list):
        return [map_nested(func, item) for item in seq]
    elif isinstance(seq, Iterator):
        return map(lambda item: map_nested(func, item), seq)
    else:
        return func(seq)


def _broadcast_trick_inner(func, shape, *args, **kwargs):
    # cupy-specific hack. numpy is happy with hardcoded shape=().
    null_shape = () if shape == () else 1

    return nxp.broadcast_to(func(*args, shape=null_shape, **kwargs), shape)


def broadcast_trick(func):
    """Apply Dask's broadcast trick to array API functions that produce arrays
    containing a single value to save space in memory.

    Note that this should only be used for arrays that never mutated.
    """
    inner = partial(_broadcast_trick_inner, func)
    inner.__doc__ = func.__doc__
    inner.__name__ = func.__name__
    return inner


# From dask.array.core, but changed to use nxp namespace
def _concatenate2(arrays, axes=None):
    """Recursively concatenate nested lists of arrays along axes

    Each entry in axes corresponds to each level of the nested list.  The
    length of axes should correspond to the level of nesting of arrays.
    If axes is an empty list or tuple, return arrays, or arrays[0] if
    arrays is a list.

    >>> x = np.array([[1, 2], [3, 4]])
    >>> _concatenate2([x, x], axes=[0])
    array([[1, 2],
           [3, 4],
           [1, 2],
           [3, 4]])

    >>> _concatenate2([x, x], axes=[1])
    array([[1, 2, 1, 2],
           [3, 4, 3, 4]])

    >>> _concatenate2([[x, x], [x, x]], axes=[0, 1])
    array([[1, 2, 1, 2],
           [3, 4, 3, 4],
           [1, 2, 1, 2],
           [3, 4, 3, 4]])

    Supports Iterators
    >>> _concatenate2(iter([x, x]), axes=[1])
    array([[1, 2, 1, 2],
           [3, 4, 3, 4]])

    Special Case
    >>> _concatenate2([x, x], axes=())
    array([[1, 2],
           [3, 4]])
    """
    if axes is None:
        axes = []

    if axes == ():
        if isinstance(arrays, list):
            return arrays[0]
        else:
            return arrays

    if isinstance(arrays, Iterator):
        arrays = list(arrays)
    if not isinstance(arrays, (list, tuple)):
        return arrays
    if len(axes) > 1:
        arrays = [_concatenate2(a, axes=axes[1:]) for a in arrays]
    concatenate = nxp.concat
    if isinstance(arrays[0], dict):
        # Handle concatenation of `dict`s, used as a replacement for structured
        # arrays when that's not supported by the array library (e.g., CuPy).
        keys = list(arrays[0].keys())
        assert all(list(a.keys()) == keys for a in arrays)
        ret = dict()
        for k in keys:
            ret[k] = concatenate(list(a[k] for a in arrays), axis=axes[0])
        return ret
    else:
        return concatenate(arrays, axis=axes[0])



import numpy as np
import numpy_groupies as npg
import pytest
from numpy.testing import assert_array_equal

import cubed.array_api as xp
from cubed.backend_array_api import namespace as nxp
from cubed.core.groupby import (
    _get_chunks_for_groups,
    groupby_blockwise,
    groupby_reduction,
)


def test_groupby_reduction_axis0():
    a = xp.full((4 * 6, 5), 7, dtype=nxp.int32, chunks=(4, 2))
    b = xp.asarray([0, 1, 0, 1] * 6, chunks=(4,))
    c = mean_groupby_reduction(a, b, axis=0, num_groups=2)
    assert_array_equal(c.compute(), np.full((2, 5), 7))


def test_groupby_reduction_axis1():
    a = xp.full((5, 4 * 6), 7, dtype=nxp.int32, chunks=(2, 4))
    b = xp.asarray([0, 1, 0, 1] * 6, chunks=(4,))
    c = mean_groupby_reduction(a, b, axis=1, num_groups=2)
    assert_array_equal(c.compute(), np.full((5, 2), 7))


def mean_groupby_reduction(x, by, axis, num_groups):
    intermediate_dtype = [("n", nxp.int64), ("total", nxp.float64)]
    dtype = x.dtype

    return groupby_reduction(
        x,
        by,
        func=_mean_groupby_func,
        combine_func=_mean_groupby_combine,
        aggregate_func=_mean_groupby_aggregate,
        axis=axis,
        intermediate_dtype=intermediate_dtype,
        dtype=dtype,
        num_groups=num_groups,
    )


def _mean_groupby_func(a, by, axis, intermediate_dtype, num_groups):
    dtype = dict(intermediate_dtype)
    n = npg.aggregate(by, a, func="len", dtype=dtype["n"], axis=axis, size=num_groups)
    total = npg.aggregate(
        by, a, func="sum", dtype=dtype["total"], axis=axis, size=num_groups
    )
    return {"n": n, "total": total}


def _mean_groupby_combine(a, axis, dummy_axis, dtype, keepdims):
    # only combine over the dummy axis, to preserve grouping along 'axis'
    dtype = dict(dtype)
    n = nxp.sum(a["n"], dtype=dtype["n"], axis=dummy_axis, keepdims=keepdims)
    total = nxp.sum(
        a["total"], dtype=dtype["total"], axis=dummy_axis, keepdims=keepdims
    )
    return {"n": n, "total": total}


def _mean_groupby_aggregate(a):
    return nxp.divide(a["total"], a["n"])


@pytest.mark.parametrize(
    "num_chunks, expected_newchunks, expected_groups_per_chunk",
    [
        [10, (3, 2, 2, 0, 3), 1],
        [5, (3, 2, 2, 0, 3), 1],
        [4, (3, 2, 2, 0, 3), 1],
        [3, (3, 2, 2, 0, 3), 1],
        [2, (5, 2, 3), 2],
        [2, (5, 2, 3), 2],
        [2, (5, 2, 3), 2],
        [2, (5, 2, 3), 2],
        [2, (5, 2, 3), 2],
        [1, (10), 5],
    ],
)
def test_get_chunks_for_groups(
    num_chunks, expected_newchunks, expected_groups_per_chunk
):
    # group 3 has no data
    labels = nxp.asarray([0, 0, 0, 1, 1, 2, 2, 4, 4, 4])
    newchunks, groups_per_chunk = _get_chunks_for_groups(
        num_chunks, labels, num_groups=5
    )
    assert_array_equal(newchunks, expected_newchunks)
    assert groups_per_chunk == expected_groups_per_chunk


def test_groupby_blockwise_axis0():
    a = xp.ones((10, 3), dtype=nxp.int32, chunks=(6, 2))
    b = nxp.asarray([0, 0, 0, 1, 1, 2, 2, 4, 4, 4])
    extra_func_kwargs = dict(dtype=nxp.int32)
    c = groupby_blockwise(
        a,
        b,
        func=_sum_reduction_func,
        axis=0,
        dtype=nxp.int64,
        num_groups=6,
        extra_func_kwargs=extra_func_kwargs,
    )
    assert_array_equal(
        c.compute(),
        nxp.asarray(
            [
                [3, 3, 3],
                [2, 2, 2],
                [2, 2, 2],
                [0, 0, 0],  # group 3 has no data
                [3, 3, 3],
                [0, 0, 0],  # final group since we specified num_groups=6
            ]
        ),
    )


def test_groupby_blockwise_axis1():
    a = xp.ones((3, 10), dtype=nxp.int32, chunks=(6, 2))
    b = nxp.asarray([0, 0, 0, 1, 1, 2, 2, 4, 4, 4])
    extra_func_kwargs = dict(dtype=nxp.int32)
    c = groupby_blockwise(
        a,
        b,
        func=_sum_reduction_func,
        axis=1,
        dtype=nxp.int64,
        num_groups=6,
        extra_func_kwargs=extra_func_kwargs,
    )
    assert_array_equal(
        c.compute(),
        nxp.asarray(
            [
                [3, 2, 2, 0, 3, 0],
                [3, 2, 2, 0, 3, 0],
                [3, 2, 2, 0, 3, 0],
            ]
        ),
    )


def _sum_reduction_func(arr, by, axis, start_group, num_groups, dtype):
    # change 'by' so it starts from 0 for each chunk
    by = by - start_group
    return npg.aggregate(by, arr, func="sum", dtype=dtype, axis=axis, size=num_groups)



import os
import time
from dataclasses import asdict
from typing import Optional

import matplotlib.patches as mpatches
import numpy as np
import pandas as pd
import pylab
import seaborn as sns

from cubed.runtime.types import Callback

sns.set_style("whitegrid")
pylab.switch_backend("Agg")


class TimelineVisualizationCallback(Callback):
    def __init__(self, format: Optional[str] = None) -> None:
        self.format = format

    def on_compute_start(self, event):
        self.start_tstamp = time.time()
        self.stats = []

    def on_task_end(self, event):
        self.stats.append(asdict(event))

    def on_compute_end(self, event):
        end_tstamp = time.time()
        dst = f"history/{event.compute_id}"
        format = self.format
        create_timeline(self.stats, self.start_tstamp, end_tstamp, dst, format)


# copy of lithops function of the same name, and modified for different field names
def create_timeline(stats, start_tstamp, end_tstamp, dst=None, format=None):
    stats_df = pd.DataFrame(stats)

    stats_df = stats_df.sort_values(by=["task_create_tstamp", "name"], ascending=True)

    total_calls = len(stats_df)

    palette = sns.color_palette("deep", 6)

    fig = pylab.figure(figsize=(10, 6))
    ax = fig.add_subplot(1, 1, 1)

    y = np.arange(total_calls)
    point_size = 10

    fields = [
        ("task create", stats_df.task_create_tstamp - start_tstamp),
        ("function start", stats_df.function_start_tstamp - start_tstamp),
        ("function end", stats_df.function_end_tstamp - start_tstamp),
        ("task result", stats_df.task_result_tstamp - start_tstamp),
    ]

    patches = []
    for f_i, (field_name, val) in enumerate(fields):
        ax.scatter(val, y, c=[palette[f_i]], edgecolor="none", s=point_size, alpha=0.8)
        patches.append(mpatches.Patch(color=palette[f_i], label=field_name))

    ax.set_xlabel("Execution Time (sec)")
    ax.set_ylabel("Function Call")

    legend = pylab.legend(handles=patches, loc="upper right", frameon=True)
    legend.get_frame().set_facecolor("#FFFFFF")

    yplot_step = int(np.max([1, total_calls / 20]))
    y_ticks = np.arange(total_calls // yplot_step + 2) * yplot_step
    ax.set_yticks(y_ticks)
    ax.set_ylim(-0.02 * total_calls, total_calls * 1.02)
    for y in y_ticks:
        ax.axhline(y, c="k", alpha=0.1, linewidth=1)

    max_seconds = np.max(end_tstamp - start_tstamp) * 1.25
    xplot_step = max(int(max_seconds / 8), 1)
    x_ticks = np.arange(max_seconds // xplot_step + 2) * xplot_step
    ax.set_xlim(0, max_seconds)

    ax.set_xticks(x_ticks)
    for x in x_ticks:
        ax.axvline(x, c="k", alpha=0.2, linewidth=0.8)

    ax.grid(False)
    fig.tight_layout()

    if format is None:
        format = "svg"

    if dst is None:
        os.makedirs("plots", exist_ok=True)
        dst = os.path.join(
            os.getcwd(), "plots", "{}_{}".format(int(time.time()), f"timeline.{format}")
        )
    else:
        dst = os.path.expanduser(dst) if "~" in dst else dst
        dst = "{}/{}".format(os.path.realpath(dst), f"timeline.{format}")

    fig.savefig(dst)



import os
from importlib import import_module

import numpy as np

# The array implementation used for backend operations is stored in the
# namespace variable, and defaults to array_api_compat.nump, unless it
# is overridden by an environment variable.
# It must be compatible with the Python Array API standard, although
# some extra functions are used too (nan functions, take_along_axis),
# which array_api_compat provides, but other Array API implementations
# may not.

if "CUBED_BACKEND_ARRAY_API_MODULE" in os.environ:
    # This code is based on similar code in array_api_tests
    xp_name = os.environ["CUBED_BACKEND_ARRAY_API_MODULE"]
    _module, _sub = xp_name, None
    if "." in xp_name:
        _module, _sub = xp_name.split(".", 1)
    xp = import_module(_module)
    if _sub:
        try:
            xp = getattr(xp, _sub)
        except AttributeError:
            # _sub may be a submodule that needs to be imported. WE can't
            # do this in every case because some array modules are not
            # submodules that can be imported (like mxnet.nd).
            xp = import_module(xp_name)
    namespace = xp

else:
    import array_api_compat.numpy

    namespace = array_api_compat.numpy


# These functions to convert to/from backend arrays
# assume that no extra memory is allocated, by using the
# Python buffer protocol.
# See https://data-apis.org/array-api/latest/API_specification/generated/array_api.asarray.html


def backend_array_to_numpy_array(arr):
    return np.asarray(arr)


def numpy_array_to_backend_array(arr, *, dtype=None):
    return namespace.asarray(arr, dtype=dtype)



from dataclasses import dataclass
from typing import Any, List, Optional

import zarr

from cubed.runtime.types import CubedPipeline
from cubed.storage.zarr import T_ZarrArray, open_if_lazy_zarr_array
from cubed.types import T_RegularChunks


@dataclass(frozen=True)
class PrimitiveOperation:
    """Encapsulates metadata about a ``blockwise`` or ``rechunk`` primitive operation."""

    pipeline: CubedPipeline
    """The pipeline that runs this operation."""

    source_array_names: List[str]
    """The names of the arrays which are inputs to this operation."""

    target_array: Any
    """The array being computed by this operation."""

    projected_mem: int
    """An upper bound of the memory needed to run a task, in bytes."""

    allowed_mem: int
    """
    The total memory available to a worker for running a task, in bytes.

    This includes any ``reserved_mem`` that has been set.
    """

    reserved_mem: int
    """The memory reserved on a worker for non-data use when running a task, in bytes."""

    num_tasks: int
    """The number of tasks needed to run this operation."""

    fusable: bool = True
    """Whether this operation should be considered for fusion."""

    write_chunks: Optional[T_RegularChunks] = None
    """The chunk size used by this operation."""


class CubedArrayProxy:
    """Generalisation of rechunker ``ArrayProxy`` with support for ``LazyZarrArray``."""

    def __init__(self, array: T_ZarrArray, chunks: T_RegularChunks):
        self.array = array
        self.chunks = chunks

    def open(self) -> zarr.Array:
        return open_if_lazy_zarr_array(self.array)


@dataclass(frozen=True)
class CubedCopySpec:
    """Generalisation of rechunker ``CopySpec`` with support for ``LazyZarrArray``."""

    read: CubedArrayProxy
    write: CubedArrayProxy


class MemoryModeller:
    """Models peak memory usage for a series of operations."""

    current_mem: int = 0
    peak_mem: int = 0

    def allocate(self, num_bytes):
        self.current_mem += num_bytes
        self.peak_mem = max(self.peak_mem, self.current_mem)

    def free(self, num_bytes):
        self.current_mem -= num_bytes
        self.peak_mem = max(self.peak_mem, self.current_mem)
